# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobufs-defined services."""
# Third Party
import grpc

# Local
from . import model_mesh_pb2 as model__mesh__pb2


class ModelMeshStub(object):
    """this is a grpc version of the external model-mesh interface
    for managing and serving models
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.registerModel = channel.unary_unary(
            "/mmesh.ModelMesh/registerModel",
            request_serializer=model__mesh__pb2.RegisterModelRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.ModelStatusInfo.FromString,
        )
        self.unregisterModel = channel.unary_unary(
            "/mmesh.ModelMesh/unregisterModel",
            request_serializer=model__mesh__pb2.UnregisterModelRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.UnregisterModelResponse.FromString,
        )
        self.getModelStatus = channel.unary_unary(
            "/mmesh.ModelMesh/getModelStatus",
            request_serializer=model__mesh__pb2.GetStatusRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.ModelStatusInfo.FromString,
        )
        self.ensureLoaded = channel.unary_unary(
            "/mmesh.ModelMesh/ensureLoaded",
            request_serializer=model__mesh__pb2.EnsureLoadedRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.ModelStatusInfo.FromString,
        )
        self.setVModel = channel.unary_unary(
            "/mmesh.ModelMesh/setVModel",
            request_serializer=model__mesh__pb2.SetVModelRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.VModelStatusInfo.FromString,
        )
        self.deleteVModel = channel.unary_unary(
            "/mmesh.ModelMesh/deleteVModel",
            request_serializer=model__mesh__pb2.DeleteVModelRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.DeleteVModelResponse.FromString,
        )
        self.getVModelStatus = channel.unary_unary(
            "/mmesh.ModelMesh/getVModelStatus",
            request_serializer=model__mesh__pb2.GetVModelStatusRequest.SerializeToString,
            response_deserializer=model__mesh__pb2.VModelStatusInfo.FromString,
        )


class ModelMeshServicer(object):
    """this is a grpc version of the external model-mesh interface
    for managing and serving models
    """

    def registerModel(self, request, context):
        """model lifecycle rpcs

        Registers a trained model to this model-mesh cluster
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def unregisterModel(self, request, context):
        """Unregisters (deletes) a model from this model-mesh cluster,
        has no effect if the specified model isn't found
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def getModelStatus(self, request, context):
        """Returns the status of the specified model. See the ModelStatus enum"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ensureLoaded(self, request, context):
        """Ensures the model with the specified id is loaded in this model-mesh cluster"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def setVModel(self, request, context):
        """vmodel lifecycle rpcs

        Creates a new vmodel id (alias) which maps to a new or existing
        concrete model, or sets the target model for an existing vmodel
        to a new or existing concrete model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def deleteVModel(self, request, context):
        """Deletes a vmodel, optionally deleting any referenced concrete
        models at the same time
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def getVModelStatus(self, request, context):
        """Gets the status of a vmodel, including associated target/active model ids
        If the vmodel is not found, the returned VModelStatusInfo will have empty
        active and target model ids and an active model status of NOT_FOUND
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ModelMeshServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "registerModel": grpc.unary_unary_rpc_method_handler(
            servicer.registerModel,
            request_deserializer=model__mesh__pb2.RegisterModelRequest.FromString,
            response_serializer=model__mesh__pb2.ModelStatusInfo.SerializeToString,
        ),
        "unregisterModel": grpc.unary_unary_rpc_method_handler(
            servicer.unregisterModel,
            request_deserializer=model__mesh__pb2.UnregisterModelRequest.FromString,
            response_serializer=model__mesh__pb2.UnregisterModelResponse.SerializeToString,
        ),
        "getModelStatus": grpc.unary_unary_rpc_method_handler(
            servicer.getModelStatus,
            request_deserializer=model__mesh__pb2.GetStatusRequest.FromString,
            response_serializer=model__mesh__pb2.ModelStatusInfo.SerializeToString,
        ),
        "ensureLoaded": grpc.unary_unary_rpc_method_handler(
            servicer.ensureLoaded,
            request_deserializer=model__mesh__pb2.EnsureLoadedRequest.FromString,
            response_serializer=model__mesh__pb2.ModelStatusInfo.SerializeToString,
        ),
        "setVModel": grpc.unary_unary_rpc_method_handler(
            servicer.setVModel,
            request_deserializer=model__mesh__pb2.SetVModelRequest.FromString,
            response_serializer=model__mesh__pb2.VModelStatusInfo.SerializeToString,
        ),
        "deleteVModel": grpc.unary_unary_rpc_method_handler(
            servicer.deleteVModel,
            request_deserializer=model__mesh__pb2.DeleteVModelRequest.FromString,
            response_serializer=model__mesh__pb2.DeleteVModelResponse.SerializeToString,
        ),
        "getVModelStatus": grpc.unary_unary_rpc_method_handler(
            servicer.getVModelStatus,
            request_deserializer=model__mesh__pb2.GetVModelStatusRequest.FromString,
            response_serializer=model__mesh__pb2.VModelStatusInfo.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "mmesh.ModelMesh", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class ModelMesh(object):
    """this is a grpc version of the external model-mesh interface
    for managing and serving models
    """

    @staticmethod
    def registerModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/registerModel",
            model__mesh__pb2.RegisterModelRequest.SerializeToString,
            model__mesh__pb2.ModelStatusInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def unregisterModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/unregisterModel",
            model__mesh__pb2.UnregisterModelRequest.SerializeToString,
            model__mesh__pb2.UnregisterModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def getModelStatus(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/getModelStatus",
            model__mesh__pb2.GetStatusRequest.SerializeToString,
            model__mesh__pb2.ModelStatusInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ensureLoaded(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/ensureLoaded",
            model__mesh__pb2.EnsureLoadedRequest.SerializeToString,
            model__mesh__pb2.ModelStatusInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def setVModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/setVModel",
            model__mesh__pb2.SetVModelRequest.SerializeToString,
            model__mesh__pb2.VModelStatusInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def deleteVModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/deleteVModel",
            model__mesh__pb2.DeleteVModelRequest.SerializeToString,
            model__mesh__pb2.DeleteVModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def getVModelStatus(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelMesh/getVModelStatus",
            model__mesh__pb2.GetVModelStatusRequest.SerializeToString,
            model__mesh__pb2.VModelStatusInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds utilities for wrapping synchronous functionality into async
wrapper functions
"""
# Standard
from concurrent.futures import ThreadPoolExecutor
from typing import AsyncGenerator, Iterable, Optional
import asyncio


def async_wrap_iter(
    it: Iterable,
    pool: Optional[ThreadPoolExecutor] = None,
) -> AsyncGenerator:
    """Wrap blocking iterable into an asynchronous one

    CITE: https://stackoverflow.com/a/62297994

    This function manages a single thread for the iteration that shuttles the
    synchronous outputs back to the async generator using a queue
    """
    loop = asyncio.get_event_loop()
    q = asyncio.Queue(1)
    exception = None
    _END = object()

    async def yield_queue_items():
        while True:
            next_item = await q.get()
            if next_item is _END:
                break
            yield next_item
        if exception is not None:
            # the iterator has raised, propagate the exception
            raise exception

    def iter_to_queue():
        nonlocal exception
        try:
            for item in it:
                # This runs outside the event loop thread, so we
                # must use thread-safe API to talk to the queue.
                asyncio.run_coroutine_threadsafe(q.put(item), loop).result()
        except Exception as e:  # pylint: disable=broad-exception-caught
            exception = e
        finally:
            asyncio.run_coroutine_threadsafe(q.put(_END), loop).result()

    loop.run_in_executor(pool, iter_to_queue)
    return yield_queue_items()



# Standard
from pathlib import Path
import json
import os
import shutil
import sys
import tempfile

# First Party
import alog

# Local
from caikit.config.config import get_config
from caikit.runtime.__main__ import main
from caikit.runtime.dump_services import dump_grpc_services, dump_http_services
import caikit

if __name__ == "__main__":
    try:
        with tempfile.TemporaryDirectory() as workdir:
            caikit.config.configure(
                config_dict={
                    "merge_strategy": "merge",
                    "runtime": {
                        "library": "sample_lib",
                        "local_models_dir": workdir,
                        "lazy_load_local_models": True,
                        "grpc": {"enabled": True},
                        "http": {"enabled": True},
                        "training": {"save_with_id": False, "output_dir": workdir},
                    },
                }
            )
            # Make sample_lib available for import
            sys.path.append(
                os.path.join(Path(__file__).parent.parent.parent, "tests/fixtures"),
            )

            # dump protos
            shutil.rmtree("protos", ignore_errors=True)
            if get_config().runtime.grpc.enabled:
                dump_grpc_services("protos")
            if get_config().runtime.http.enabled:
                dump_http_services("protos")

            # create a sample.json file for training
            with open(
                os.path.join("protos", "sample.json"), "w", encoding="utf-8"
            ) as handle:
                handle.write(
                    json.dumps(
                        [
                            {"number": 1},
                            {"number": 2},
                        ]
                    )
                )

            alog.configure(default_level="debug")
            main()
    finally:
        # remove generated protos
        shutil.rmtree("protos", ignore_errors=True)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import json

# Third Party
import grpc
import requests

# Local
from caikit.config.config import get_config
from caikit.runtime import get_inference_request
from caikit.runtime.service_factory import ServicePackageFactory
from text_sentiment.runtime_model import HuggingFaceSentimentModule
import caikit

if __name__ == "__main__":
    caikit.config.configure(
        config_dict={
            "merge_strategy": "merge",
            "runtime": {
                "library": "text_sentiment",
                "grpc": {"enabled": True},
                "http": {"enabled": True},
            },
        }
    )

    inference_service = ServicePackageFactory().get_service_package(
        ServicePackageFactory.ServiceType.INFERENCE,
    )

    model_id = "text_sentiment"

    if get_config().runtime.grpc.enabled:
        # Setup the client
        port = 8085
        channel = grpc.insecure_channel(f"localhost:{port}")
        client_stub = inference_service.stub_class(channel)

        # Run inference for two sample prompts
        for text in ["I am not feeling well today!", "Today is a nice sunny day"]:
            request = get_inference_request(HuggingFaceSentimentModule.TASK_CLASS)(
                text_input=text
            ).to_proto()
            response = client_stub.HuggingFaceSentimentTaskPredict(
                request, metadata=[("mm-model-id", model_id)], timeout=1
            )
            print("Text:", text)
            print("RESPONSE from gRPC:", response)

    if get_config().runtime.http.enabled:
        port = 8080
        # Run inference for two sample prompts
        for text in ["I am not feeling well today!", "Today is a nice sunny day"]:
            payload = {"inputs": text}
            response = requests.post(
                f"http://localhost:{port}/api/v1/{model_id}/task/hugging-face-sentiment",
                json=payload,
                timeout=1,
            )
            print("\nText:", text)
            print("RESPONSE from HTTP:", json.dumps(response.json(), indent=4))



"""
Dummy data model object for testing
"""
# Standard
from typing import Iterable, Union
import base64
import typing

# Local
from caikit.core import DataObjectBase, TaskBase, dataobject, task
from caikit.core.data_model import ProducerId


@dataobject(package="caikit_data_model.sample_lib")
class SampleInputType(DataObjectBase):
    """A sample `domain primitive` input type for this library.
    The analog to a `Raw Document` for the `Natural Language Processing` domain."""

    name: str


@dataobject(package="caikit_data_model.sample_lib")
class SampleOutputType(DataObjectBase):
    """A simple return type for the `sample_task` task"""

    greeting: str


@dataobject(package="caikit_data_model.sample_lib")
class OtherOutputType(DataObjectBase):
    """A simple return type for the `other_task` task"""

    farewell: str
    producer_id: ProducerId


@dataobject(package="caikit_data_model.sample_lib")
class FileDataType(DataObjectBase):
    """A simple type for tasks that deal with file data"""

    filename: str
    data: bytes


@dataobject(package="caikit_data_model.sample_lib")
class SampleTrainingType(DataObjectBase):
    """A sample `training data` type for the `sample_task` task."""

    number: int
    label: str


@task(
    unary_parameters={"sample_input": SampleInputType},
    streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
    unary_output_type=SampleOutputType,
    streaming_output_type=Iterable[SampleOutputType],
)
class SampleTask(TaskBase):
    """A sample `task` for our test models"""


@task(
    unary_parameters={"sample_input": SampleInputType},
    unary_output_type=OtherOutputType,
)
class OtherTask(TaskBase):
    """Another sample `task` for our test models"""


@task(
    unary_parameters={"unprocessed": FileDataType},
    unary_output_type=FileDataType,
)
class FileTask(TaskBase):
    """A sample task for processing files"""


@task(
    streaming_parameters={"lats": Iterable[float], "lons": Iterable[float]},
    streaming_output_type=Iterable[SampleOutputType],
)
class GeoSpatialTask(TaskBase):
    """A task that flexes streaming capabilities"""


# NB: Backwards compatibility test
@task(
    required_parameters={"sample_input": SampleInputType},
    output_type=typing.Iterable[SampleOutputType],
)
class StreamingTask(TaskBase):
    """A streaming version of a task"""



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains interfaces required to generate DataStreamSource[T] classes
"""

# Standard
from typing import List

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber

# Local
from caikit.core.data_model import PACKAGE_COMMON, DataObjectBase, dataobject


@dataobject(PACKAGE_COMMON)
class File(DataObjectBase):
    filename: Annotated[str, FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class ListOfFiles(DataObjectBase):
    files: Annotated[List[str], FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class Directory(DataObjectBase):
    dirname: Annotated[str, FieldNumber(1)]
    extension: Annotated[str, FieldNumber(2)]


@dataobject(PACKAGE_COMMON)
class S3Base(DataObjectBase):
    # URI info
    endpoint: Annotated[str, FieldNumber(2)]  # begins with `http://` or `https://`
    region: Annotated[str, FieldNumber(3)]
    bucket: Annotated[str, FieldNumber(4)]

    # HMAC credentials
    accessKey: Annotated[str, FieldNumber(5)]
    secretKey: Annotated[str, FieldNumber(6)]

    # IAM credentials
    IAM_id: Annotated[str, FieldNumber(7)]
    IAM_api_key: Annotated[str, FieldNumber(8)]


@dataobject(PACKAGE_COMMON)
class S3Files(S3Base):
    # List of file paths relative to the bucket
    files: Annotated[List[str], FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class S3Path(S3Base):
    # Path relative to the bucket
    path: Annotated[str, FieldNumber(1)]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A DestroyableProcess implements a multiprocessing Process that captures errors
and communicates them to its parent with clean semantics for being destroyed.

NOTE: The "fork" start method is used for two reasons:
    1. It's faster
    2. For auto-generated classes (e.g. stream sources), "spawn" can result in
        missing classes since it performs a full re-import, but may not
        regenerate these classes.
"""
# Standard
from functools import partial
from typing import Any, Callable, Optional, Tuple
import multiprocessing
import os

# First Party
import alog

# Local
from ..exceptions import error_handler
from .destroyable import Destroyable

log = alog.use_channel("DESTROY-PROC")
error = error_handler.get(log)


OOM_EXIT_CODES = [137, 9, -9]

FORK_CTX = multiprocessing.get_context("fork")
SPAWN_CTX = multiprocessing.get_context("spawn")
FORKSERVER_CTX = multiprocessing.get_context("forkserver")


class _DestroyableProcess(
    multiprocessing.process.BaseProcess,
    Destroyable,
):  # pylint: disable=too-many-instance-attributes
    """The _DestroyableProcess base class implements a context-agnostic process
    class that manages the subprocess and allows it to be destroyed
    """

    def __init__(
        self,
        target: Optional[Callable] = None,
        completion_event: Optional[multiprocessing.Event] = None,
        args: Optional[Tuple] = None,
        kwargs: Optional[dict] = None,
        destroy_grace_period: float = 10,
        return_result: bool = False,
        **_kwargs,
    ):
        """Initialize with an event to use to signal completion"""
        self._parent_conn, self._child_conn = self.__class__._MP_CTX.Pipe()
        self._completion_event = completion_event or self.__class__._MP_CTX.Event()
        self._destroy_grace_period = destroy_grace_period
        self._return_result = return_result

        # This will hold the terminal result of the process which will either be
        # an Exception or some other value
        self.__result = None

        # Descriptions of completion state
        self.__started = False
        self.__destroyed = False
        self.__canceled = False

        # Bind the target to the target wrapper
        wrapped_target = partial(
            self._target_wrapper,
            target=target,
            args=args,
            kwargs=kwargs,
        )
        super().__init__(target=wrapped_target, **_kwargs)

    ## Destroyable Interface ##

    @property
    def destroyed(self) -> bool:
        return self.__destroyed

    @property
    def canceled(self) -> bool:
        return self.__canceled

    @property
    def ran(self) -> bool:
        return self.__started and not self.is_alive()

    @property
    def threw(self) -> bool:
        return self.error is not None

    def get_or_throw(self) -> Any:
        """Get the result of the execution or raise an error if one occurred"""
        if self.destroyed:
            log.error(
                "<COR24981767E>",
                "get_or_throw called on destroyed process, no value to return",
            )

        if not self.ran:
            log.error(
                "<COR12037430E>",
                "get_or_throw called on process, but it has not finished running",
            )

        # Update the result and throw if it's an error
        err = self.error
        if err is not None:
            raise err
        return self.__result

    def destroy(self):
        """Cancel any in-progress work"""
        self.__destroyed = True
        if self.is_alive() or not self.__started:
            self.__canceled = True
        if self.__started:
            self.terminate()
            self.join(self._destroy_grace_period)
            self.kill()
            self.join()
            self._completion_event.set()

    ## Process Interface ##

    def start(self):
        if self.destroyed:
            err_msg = "Not starting work on pre-canceled process"
            log.warning("<COR42191929W>", err_msg)
            self._completion_event.set()
            self.__result = RuntimeError(err_msg)
            return
        self.__started = True
        return super().start()

    def join(self, *args, **kwargs):
        if self.destroyed and not self.__started:
            return
        return super().join(*args, **kwargs)

    # NOTE: This functionality is covered by the unit tests, but pytest-cov is
    #   not correctly collecting the coverage info since it executes inside the
    #   subprocess
    def run(self):  # pragma: no cover
        try:
            # Run and indicate to the parent that no
            super().run()

        # Catch any errors thrown within a subprocess so that they can be
        # forwarded to the parent
        # pylint: disable=broad-exception-caught
        except Exception as err:
            err_str = repr(err)
            log.error(
                "<COR69863806E>",
                "Caught exception in destroyable process: %s",
                err_str,
                exc_info=True,
            )
            self._child_conn.send(err)
        finally:
            self._completion_event.set()

    ## Impl ##

    def _update_result(self):
        if self._parent_conn.poll():
            self.__result = self._parent_conn.recv()

    @property
    def error(self) -> Optional[Exception]:
        self._update_result()

        if isinstance(self.__result, Exception):
            return self.__result

        if self.exitcode and self.exitcode != os.EX_OK:
            if self.exitcode in OOM_EXIT_CODES:
                return MemoryError("Training process died with OOM error!")
            if not self.canceled:
                return RuntimeError(
                    f"Training process died with exit code {self.exitcode}"
                )

    @property
    def completion_event(self) -> multiprocessing.Event:
        return self._completion_event

    # NOTE: This functionality is covered by the unit tests, but pytest-cov is
    #   not correctly collecting the coverage info since it executes inside the
    #   subprocess
    def _target_wrapper(self, target, args, kwargs):  # pragma: no cover
        result = target(*(args or []), **(kwargs or {}))
        log.debug3("Process target result: %s", result)
        if self._return_result:
            self._child_conn.send(result)
        self._completion_event.set()


class _ForkDestroyableProcess(FORK_CTX.Process, _DestroyableProcess):
    _MP_CTX = FORK_CTX


class _SpawnDestroyableProcess(SPAWN_CTX.Process, _DestroyableProcess):
    _MP_CTX = SPAWN_CTX


class _ForkserverDestroyableProcess(FORKSERVER_CTX.Process, _DestroyableProcess):
    _MP_CTX = FORKSERVER_CTX


_PROCESS_TYPES = {
    "fork": _ForkDestroyableProcess,
    "forkserver": _ForkserverDestroyableProcess,
    "spawn": _SpawnDestroyableProcess,
}


def DestroyableProcess(start_method: str, *args, **kwargs):
    """Class wrapper that returns the appropriate process type based on the
    requested start method.

    NOTE: Naming intentionally looks like a class!
    """
    error.value_check(
        "<COR16699811E>",
        start_method in _PROCESS_TYPES,
        "Unsupported start_method: {}",
        start_method,
    )
    return _PROCESS_TYPES[start_method](*args, **kwargs)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from os import getenv, path
import asyncio
import subprocess

# Third Party
import pytest

# Local
from tests.examples.shared import requirements, waitForPort


@pytest.mark.examples
def test_example_text_sentiment():
    # Example specific grpc port
    grpc_port = 8085

    with requirements("text-sentiment") as (python_venv, example_dir):
        # Start the server
        with subprocess.Popen(
            [python_venv, "start_runtime.py"],
            cwd=example_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as server:
            # Check if the gRPC port is open
            # The gRPC server start-up time has some inherent variability
            # 60s timeout should cover most situations, while keeping the
            # test execution time reasonable
            if not asyncio.run(waitForPort(grpc_port, 60)):
                server.terminate()
                pytest.fail(
                    "Failed to connect to the gRPC server on port {} in 30s.".format(
                        grpc_port
                    )
                )

            # Server is running, start the client
            # Use a timeout of 10s for inference. Capture outputs to report
            # them in case of failure.
            try:
                subprocess.run(
                    [python_venv, path.join(example_dir, "client.py")],
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    timeout=30,
                )
            except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
                server.terminate()
                pytest.fail("Client failed with output: {}".format(e))

            # Client worked well, let's stop the server
            server.terminate()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The dict-based backend implementation
"""

# Standard
from typing import Any, Iterable, Type

# First Party
import alog

# Local
from ...exceptions import error_handler
from ..base import DataBase
from .base import DataModelBackendBase

log = alog.use_channel("DATAB")
error = error_handler.get(log)


# DictBackend ##################################################################


class DictBackend(DataModelBackendBase):
    """Data model backend for a raw dict"""

    def __init__(self, data_dict: dict):
        """Construct with the dict"""
        error.type_check("<COR85037210E>", dict, data_dict=data_dict)
        self._data_dict = data_dict

    def get_attribute(self, data_model_class: Type[DataBase], name: str) -> Any:
        """Fetch the attribute out of the internal dict and validate it against
        the target data model class. If the target attribute is a nested data
        model type, wrap the corresponding nested dict in an instance of this
        same backend.

        Args:
            data_model_class (Type[DataBase]): The frontend data model class
                that is accessing this attribute
            name (str): The name of the attribute to access

        Returns:
            value:  Any
                The extracted attribute value
        """

        # NOTE: We do not type-check the args here for efficiency. This method
        #   should only be called by the DataBase class, so we can assume it's
        #   being used correctly.

        # Make sure the name is a valid field on the given class
        if (
            name not in data_model_class.fields
            and name not in data_model_class._fields_oneofs_map
        ):
            error(
                "<COR85037211E>",
                AttributeError(
                    f"No such attribute [{name}] on [{data_model_class.__name__}]"
                ),
            )

        # Get the value from the internal dict
        raw_value = self._data_dict.get(name)

        # If the target attribute is itself a message, make sure the value is a
        # dict, then wrap it in the corresponding data model object with the a
        # new backend instance.
        if name in data_model_class._fields_message and raw_value is not None:
            error.type_check("<COR85037212E>", dict, **{name: raw_value})
            proto_class = data_model_class.get_proto_class()
            field_dm_class = DataBase.get_class_for_proto(
                proto_class.DESCRIPTOR.fields_by_name[name].message_type
            )
            return field_dm_class.from_backend(self.__class__(raw_value))

        # If the target attribute is a repeated message, convert it to a list of
        # nested messages with dict backends
        #
        # TODO: It may be better to do this lazily, but that would come at the
        #   expense of being able to re-iterate or randomly-index into the
        #   object. We could consider writing a lazily constructed list to avoid
        #   costruction of the messages that aren't used.
        if name in data_model_class._fields_message_repeated and raw_value is not None:
            error.type_check("<COR85037213E>", Iterable, **{name: raw_value})
            field_dm_class = data_model_class.get_field_message_type(name)
            return [
                field_dm_class.from_backend(self.__class__(entry))
                for entry in raw_value
            ]

        # If the target attribure is a oneof name, look for the oneof name in the data dict first.
        # If it exists, infer the right field name based on the type of the value and return the
        # field name and the value. If oneof name isn't in the data dict, check for field name
        # instead and return the right value
        if name in data_model_class._fields_oneofs_map:
            if name in self._data_dict:
                val = self._data_dict[name]
                which_oneof = data_model_class._infer_which_oneof(name, val)
                return data_model_class.OneofFieldVal(val=val, which_oneof=which_oneof)

            oneof_fields = data_model_class._fields_oneofs_map[name]
            for field in oneof_fields:
                if field in self._data_dict:
                    return data_model_class.OneofFieldVal(
                        val=self._data_dict[field], which_oneof=field
                    )

        return raw_value



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module holds the Pydantic wrapping required by the REST server,
capable of converting to and from Pydantic models to our DataObjects.
"""
# Standard
from typing import Dict, List, Type, Union, get_args, get_type_hints
import base64
import enum

# Third Party
from pydantic.functional_validators import BeforeValidator
import numpy as np
import pydantic

# First Party
from py_to_proto.dataclass_to_proto import (  # Imported here for 3.8 compat
    Annotated,
    get_origin,
)

# Local
from caikit.core.data_model.base import DataBase
from caikit.interfaces.common.data_model.primitive_sequences import (
    BoolSequence,
    FloatSequence,
    IntSequence,
    StrSequence,
)

# PYDANTIC_TO_DM_MAPPING is essentially a 2-way map of DMs <-> Pydantic models, you give it a
# pydantic model, it gives you back a DM class, you give it a
# DM class, you get back a pydantic model.
PYDANTIC_TO_DM_MAPPING = {
    # Map primitive sequences to lists
    StrSequence: List[str],
    IntSequence: List[int],
    FloatSequence: List[float],
    BoolSequence: List[bool],
}


# Base class for pydantic models
# We want to set the config to forbid extra attributes
# while instantiating any pydantic models
# This is done to make sure any oneofs can be
# correctly inferred by pydantic
class ParentPydanticBaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", protected_namespaces=())


def pydantic_to_dataobject(pydantic_model: pydantic.BaseModel) -> DataBase:
    """Convert pydantic objects to our DM objects"""
    dm_class_to_build = PYDANTIC_TO_DM_MAPPING.get(type(pydantic_model))
    dm_kwargs = {}

    for field_name, field_value in pydantic_model:
        # field could be a DM:
        # pylint: disable=unidiomatic-typecheck
        if type(field_value) in PYDANTIC_TO_DM_MAPPING:
            dm_kwargs[field_name] = pydantic_to_dataobject(field_value)
        elif isinstance(field_value, list):
            if all(type(val) in PYDANTIC_TO_DM_MAPPING for val in field_value):
                dm_kwargs[field_name] = [
                    pydantic_to_dataobject(val) for val in field_value
                ]
            else:
                dm_kwargs[field_name] = field_value
        else:
            dm_kwargs[field_name] = field_value

    return dm_class_to_build(**dm_kwargs)


def dataobject_to_pydantic(dm_class: Type[DataBase]) -> Type[pydantic.BaseModel]:
    """Make a pydantic model based on the given proto message by using the data
    model class annotations to mirror as a pydantic model
    """
    # define a local namespace for type hints to get type information from.
    # This is needed for pydantic to have a handle on JsonDict and JsonDictValue while
    # creating its base model
    localns = {"JsonDict": dict, "JsonDictValue": dict}

    if dm_class in PYDANTIC_TO_DM_MAPPING:
        return PYDANTIC_TO_DM_MAPPING[dm_class]

    annotations = {
        field_name: _get_pydantic_type(field_type)
        for field_name, field_type in get_type_hints(dm_class, localns=localns).items()
    }
    pydantic_model = type(ParentPydanticBaseModel)(
        dm_class.get_proto_class().DESCRIPTOR.full_name,
        (ParentPydanticBaseModel,),
        {
            "__annotations__": annotations,
            **{
                name: None
                for name, _ in get_type_hints(
                    dm_class,
                    localns=localns,
                ).items()
            },
        },
    )
    PYDANTIC_TO_DM_MAPPING[dm_class] = pydantic_model
    # also store the reverse mapping for easy retrieval
    # should be fine since we only check for dm_class in this dict
    PYDANTIC_TO_DM_MAPPING[pydantic_model] = dm_class
    return pydantic_model


# pylint: disable=too-many-return-statements
def _get_pydantic_type(field_type: type) -> type:
    """Recursive helper to get a valid pydantic type for every field type"""
    # pylint: disable=too-many-return-statements

    # Leaves: we should have primitive types and enums
    if np.issubclass_(field_type, np.integer):
        return int
    if np.issubclass_(field_type, np.floating):
        return float
    if field_type == bytes:
        return Annotated[bytes, BeforeValidator(_from_base64)]
    if field_type in (int, float, bool, str, dict, type(None)):
        return field_type
    if isinstance(field_type, type) and issubclass(field_type, enum.Enum):
        return field_type

    # These can be nested within other data models
    if (
        isinstance(field_type, type)
        and issubclass(field_type, DataBase)
        and not issubclass(field_type, pydantic.BaseModel)
    ):
        # NB: for data models we're calling the data model conversion fn
        return dataobject_to_pydantic(field_type)

    # And then all of these types can be nested in other type annotations
    if get_origin(field_type) is Annotated:
        return _get_pydantic_type(get_args(field_type)[0])
    if get_origin(field_type) is Union:
        return Union[  # type: ignore
            tuple((_get_pydantic_type(arg_type) for arg_type in get_args(field_type)))
        ]
    if get_origin(field_type) is list:
        return List[_get_pydantic_type(get_args(field_type)[0])]

    if get_origin(field_type) is dict:
        return Dict[
            _get_pydantic_type(get_args(field_type)[0]),
            _get_pydantic_type(get_args(field_type)[1]),
        ]

    raise TypeError(f"Cannot get pydantic type for type [{field_type}]")


def _from_base64(data: Union[bytes, str]) -> bytes:
    if isinstance(data, str):
        return base64.b64decode(data.encode("utf-8"))
    return data



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""DataStream converter that attempts to consolidate some columns from a CSV.
Useful for both collapsing multiple columns of a CSV stream into one list, or creating a list out of
a single column from a CSV file.
"""
# Standard
from typing import Dict

# First Party
import alog

# Local
from ...exceptions import error_handler
from .. import DataStream

log = alog.use_channel("CSVCLMNCNSLDTR")
error = error_handler.get(log)


class CSVColumnFormatter:
    """Consolidates column(s) of data items in a csv stream.
    Only operates on streams of lists.

    For example, for an input stream that looks like:
        [
            ["foo", "bar"],
            ["bazz", "buzz", "bozz"]
        ]
    If the expected_columns are {'text': str, 'labels': list}, the result will be:
        [
            ["foo", ["bar"]],
            ["bazz", ["buzz, "bozz"]]
        ]

    >>> list_of_lists = [["foo", "bar"], ["bazz", "buzz", "bozz"]]
    >>> list_stream = DataStream.from_iterable(list_of_lists)
    >>> converter = CSVColumnFormatter(expected_columns={'text': str, 'labels': list})
    >>> list_stream = converter.format(list_stream)
    >>> # [ ["foo", ["bar"]], ["bazz", ["buzz, "bozz"]] ]
    """

    def __init__(self, expected_columns: Dict[str, type]):
        """Initialize CSVColumnConsolidator

        Args:
            expected_columns (Dict(str, type)): (Ordered) dictionary mapping the
                csv column names to expected types
        """
        error.type_check("<COR56775937E>", dict, expected_columns=expected_columns)
        error.value_check(
            "<COR98237989E>", expected_columns, "`expected_columns` should be nonempty"
        )
        self._expected_columns: Dict[str, type] = expected_columns

    def format(self, stream: DataStream) -> DataStream:
        """Attempt to format some CSV columns. Only operates on streams of lists.

        Pulls column elements into lists if the column type is `list`
        Slurps remaining columns into the last column if the last column type is `list`

        See classdoc for examples

        Args:
            stream (DataStream): stream intended to be converted

        Returns:
            DataStream:
                stream with internal elements of each item converted
        """

        # Element listification only works for lists.
        # If neither is asked for, do nothing.
        if list not in self._expected_columns.values():
            return stream

        def _convert(data_item):
            if isinstance(data_item, list):
                # Don't mutate a list that we're iterating on here
                # Subsequent re-entries into the stream would mutate the list further
                # and really mess things up
                data_item_copy = list(data_item[:])

                last_type = None
                for i, (element, type_) in enumerate(
                    zip(data_item, self._expected_columns.values())
                ):
                    last_type = type_
                    if type_ == list:
                        data_item_copy[i] = CSVColumnFormatter._attempt_to_listify(
                            element
                        )

                if len(data_item) > len(self._expected_columns):
                    # More data in the data item left...
                    if last_type == list:
                        # Last element was a list, so slurp the rest of the row in
                        length = len(self._expected_columns)

                        data_item_copy[length - 1].extend(data_item[length:])
                        data_item_copy = data_item_copy[0:length]

                return data_item_copy

            # Do nothing if thing wasn't a list
            return data_item

        return stream.map(_convert)

    @staticmethod
    def _attempt_to_listify(data):
        """Mostly we want to return [data] if this thing isn't already a list.
        But we don't want to do things like convert a dictionary into a list of keys.

        So we'll just check for some common types, and listify those.
        """
        if isinstance(data, (bool, float, int, str)):
            return [data]
        return data



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Have pylint ignore Class XXXX has no YYYY member so that we can use gRPC enums.
# pylint: disable=E1101

# First Party
import alog

# Local
from caikit import get_config
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.protobufs import model_runtime_pb2, model_runtime_pb2_grpc
from caikit.runtime.types.aborted_exception import AbortedException
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.work_management.abortable_action import AbortableAction
from caikit.runtime.work_management.rpc_aborter import RpcAborter

log = alog.use_channel("MR-SERVICR-I")


class ModelRuntimeServicerImpl(model_runtime_pb2_grpc.ModelRuntimeServicer):
    """This class contains the implementation of all of the RPCs that are required to run a
    service in Model Mesh as a Model-Runtime."""

    def __init__(self):
        self.model_manager = ModelManager.get_instance()

    def loadModel(self, request, context):
        """Model loading .
        Args:
            request(model_runtime_pb2.LoadModelRequest):
                gRPC request, gen from model-runtime.proto
            context(grpc.ServicerContext):
                Context object (contains request metadata, etc)
        Returns:
            model_runtime_pb2.LoadModelResponse: Gen from model-runtime.proto
        """
        try:
            log.info(
                {
                    "log_code": "<RUN10000106I>",
                    "message": "Loading model '%s'" % request.modelId,
                    "model_id": request.modelId,
                }
            )
            caikit_config = get_config()
            if caikit_config.runtime.use_abortable_threads:
                aborter = RpcAborter(context)
                work = AbortableAction(
                    aborter,
                    self.model_manager.load_model,
                    request.modelId,
                    request.modelPath,
                    request.modelType,
                    aborter=aborter,
                )
                model_size = work.do()
            else:
                model_size = self.model_manager.load_model(
                    request.modelId, request.modelPath, request.modelType
                )

            log.info(
                {
                    "log_code": "<RUN10000107I>",
                    "message": "Model '%s' loaded! Model size [%s]"
                    % (request.modelId, str(model_size)),
                    "model_id": request.modelId,
                }
            )

        except AbortedException as e:
            log.warning(
                {
                    "log_code": "<RUN82590507W>",
                    "message": "Model '%s' was not loaded due to the rpc aborting"
                    % request.modelId,
                    "model_id": request.modelId,
                    "error_id": e.id,
                }
            )

            # Unload the model in case it had actually finished loading
            self.model_manager.unload_model(request.modelId)

            raise e

        except CaikitRuntimeException as e:
            log.warning(
                {
                    "log_code": "<RUN84720101W>",
                    "message": "Model '%s' could not be loaded! Reason: [%s]"
                    % (request.modelId, str(e.message)),
                    "model_id": request.modelId,
                    "error_id": e.id,
                }
            )
            raise e

        # get concurrency
        model_mesh_config = get_config().inference_plugin.model_mesh
        if request.modelType in model_mesh_config.max_model_concurrency_per_type:
            max_concurrency = model_mesh_config.max_model_concurrency_per_type[
                request.modelType
            ]
        else:
            max_concurrency = model_mesh_config.max_model_concurrency

        return model_runtime_pb2.LoadModelResponse(
            sizeInBytes=model_size, maxConcurrency=max_concurrency
        )

    def unloadModel(self, request, context):
        """Model unloading.

        Args:
            request(model_runtime_pb2.UnloadModelRequest):
                gRPC request, gen from model-runtime.proto
            context(grpc.ServicerContext):
                Context object (contains request metadata, etc)
        Returns:
            model_runtime_pb2.UnloadModelResponse: Gen from model-runtime.proto
        """
        try:
            log.info(
                {
                    "log_code": "<RUN10000110I>",
                    "message": "Unloading model '%s'" % request.modelId,
                    "model_id": request.modelId,
                }
            )
            model_size = self.model_manager.unload_model(request.modelId)
            log.info(
                {
                    "log_code": "<RUN10000111I>",
                    "message": "Unloaded model '%s' (Reclaimed size: %s)"
                    % (request.modelId, model_size),
                    "model_id": request.modelId,
                }
            )
        except CaikitRuntimeException as e:
            log.warning(
                {
                    "log_code": "<RUN18471838W>",
                    "message": "Model '%s' could not be unloaded! Reason: [%s]"
                    % (request.modelId, str(e.message)),
                    "model_id": request.modelId,
                    "error_id": e.id,
                }
            )
            raise e
        return model_runtime_pb2.UnloadModelResponse()

    def predictModelSize(self, request, context):
        """Predict size of not-yet-loaded model

        Args:
            request(model_runtime_pb2.PredictModelSizeRequest):
                gRPC request, gen from model-runtime.proto
            context(grpc._server._Context):
                Context object (contains request metadata, etc)
        Returns:
            model_runtime_pb2.PredictModelSizeResponse: Gen from model-runtime.proto
        """
        try:
            log.info(
                {
                    "log_code": "<RUN10000120I>",
                    "message": "Predicting size of model '%s'" % request.modelId,
                    "model_id": request.modelId,
                }
            )
            predicted_size = self.model_manager.estimate_model_size(
                request.modelId, request.modelPath, request.modelType
            )
            log.info(
                {
                    "log_code": "<RUN10000123I>",
                    "message": "Predicted model '%s' size: [%s]"
                    % (request.modelId, str(predicted_size)),
                    "model_id": request.modelId,
                }
            )

        except CaikitRuntimeException as e:
            log.warning(
                {
                    "log_code": "<RUN14920102W>",
                    "message": "Model '%s' size could not be predicted! Reason: [%s]"
                    % (request.modelId, e.message),
                    "model_id": request.modelId,
                    "error_id": e.id,
                }
            )
            raise e
        return model_runtime_pb2.PredictModelSizeResponse(sizeInBytes=predicted_size)

    def modelSize(self, request, context):
        """Compute size (memory consumption) of currently-loaded model

        Args:
            request(model_runtime_pb2.ModelSizeRequest):
                gRPC request, gen from model-runtime.proto
            context(grpc._server._Context):
                Context object (contains request metadata, etc)
        Returns:
            model_runtime_pb2.ModelSizeResponse: Gen from model-runtime.proto
        """
        try:
            log.info(
                {
                    "log_code": "<RUN10000121I>",
                    "message": "Computing size of model '%s'" % request.modelId,
                    "model_id": request.modelId,
                }
            )
            model_size = self.model_manager.get_model_size(request.modelId)
            log.info(
                {
                    "log_code": "<RUN10000122I>",
                    "message": "Computed model '%s' size: [%s]"
                    % (request.modelId, str(model_size)),
                    "model_id": request.modelId,
                }
            )
        except CaikitRuntimeException as e:
            log.warning(
                {
                    "log_code": "<RUN14440122W>",
                    "message": "Failed to calculate model '%s' size! Reason: [%s]"
                    % (request.modelId, e.message),
                    "model_id": request.modelId,
                    "error_id": e.id,
                }
            )
            raise e
        return model_runtime_pb2.ModelSizeResponse(sizeInBytes=model_size)

    def runtimeStatus(self, request, context):
        """Runtime status checking.

        Args:
            request(model_runtime_pb2.RuntimeStatusRequest):
                GRPC request, gen from model-runtime.proto
            context(grpc.ServicerContext):
                Context object (contains request metadata, etc)
        Returns:
            model_runtime_pb2.RuntimeStatusResponse:
                Gen from model-runtime.proto
        """
        model_mesh_config = get_config().inference_plugin.model_mesh
        log.info(
            "<RUN25209721I>",
            "Starting Model Runtime version: %s",
            model_mesh_config.runtime_version,
        )
        return model_runtime_pb2.RuntimeStatusResponse(
            status=model_runtime_pb2.RuntimeStatusResponse.READY,
            capacityInBytes=model_mesh_config.capacity,
            maxLoadingConcurrency=model_mesh_config.max_loading_concurrency,
            modelLoadingTimeoutMs=model_mesh_config.model_loading_timeout_ms,
            defaultModelSizeInBytes=model_mesh_config.default_model_size,
            runtimeVersion=model_mesh_config.runtime_version,
            numericRuntimeVersion=model_mesh_config.numeric_runtime_version,
            limitModelConcurrency=model_mesh_config.latency_based_autoscaling_enabled,
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains logic to ensure proto-API compatibility and model support compatibility"""

# Standard
from typing import Dict

# First Party
import alog

log = alog.use_channel("CMPAT-CHKR")


class ApiFieldNames:
    """
    Singleton class for fetching API spec from service module
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
            cls._instance.service_pb2_modules = []
        return cls._instance

    @staticmethod
    def clear():
        ApiFieldNames().service_pb2_modules = []

    @staticmethod
    def get_fields_for_message(message_name: str) -> Dict[str, int]:
        """Return first matching message found for 'message_name'"""
        for service in ApiFieldNames().service_pb2_modules:
            if message_name in service.DESCRIPTOR.message_types_by_name:
                message_type = service.DESCRIPTOR.message_types_by_name[message_name]

                message_fields = message_type.fields
                return {field.name: field.number for field in message_fields}

            log.info("Did not find message name in proto module: %s", message_name)
        return {}



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for LocalModelTrainer
"""
# Standard
from datetime import timedelta
import multiprocessing
import os
import tempfile
import threading

# Third Party
import pytest

# First Party
import aconfig

# Local
from caikit.config import get_config
from caikit.core.data_model import DataStream, TrainingStatus
from caikit.core.exceptions.caikit_core_exception import CaikitCoreException
from caikit.core.model_management.local_model_trainer import LocalModelTrainer
from sample_lib.modules import SampleModule

## Helpers #####################################################################


def local_trainer(**kwargs) -> LocalModelTrainer:
    cfg = aconfig.Config(kwargs, override_env_vars=False)
    return LocalModelTrainer(cfg, "test-instance")


@pytest.fixture(params=[True, False])
def trainer_type_cfg(request):
    yield {"use_subprocess": request.param}


@pytest.fixture
def save_path():
    with tempfile.TemporaryDirectory() as workdir:
        yield os.path.join(workdir, "model_save_dir")


def get_event(cfg: dict):
    if cfg.get("use_subprocess"):
        start_method = (
            get_config().model_management.trainers.default.config.subprocess_start_method
        )
        return multiprocessing.get_context(start_method).Event()
    return threading.Event()


## Tests #######################################################################


def test_train_and_get_info(trainer_type_cfg):
    """Test that running a training can fetch status correctly"""
    trainer = local_trainer(**trainer_type_cfg)

    # Launch the training and force it to wait
    # NOTE: Data stream passed by positional arg to ensure it is passed through
    #   correctly by position
    wait_event = get_event(trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        DataStream.from_iterable([]),
        wait_event=wait_event,
    )
    assert model_future.get_info().status == TrainingStatus.RUNNING
    assert not model_future.get_info().status.is_terminal

    # Let the training proceed and wait for it to complete
    wait_event.set()
    model_future.wait()
    assert model_future.get_info().status == TrainingStatus.COMPLETED
    assert model_future.get_info().status.is_terminal

    # Re-fetch the future by ID
    fetched_future = trainer.get_model_future(model_future.id)
    assert fetched_future is model_future


def test_train_save_and_load(trainer_type_cfg, save_path):
    """Test that a trained model can be loaded"""
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        save_path=save_path,
    )
    model_future.wait()

    # Make sure it can be loaded manually
    model = SampleModule.load(save_path)
    assert isinstance(model, SampleModule)

    # Make sure that it can be loaded via the future
    # NOTE: With a subprocess, this requires that save path is given so that the
    #   model can be re-loaded from disk
    model = model_future.load()
    assert isinstance(model, SampleModule)


def test_save_with_id(trainer_type_cfg, save_path):
    """Test that saving with the training id correctly injects the ID in the
    save path
    """
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        save_path=save_path,
        save_with_id=True,
    )
    model_future.wait()
    assert model_future.save_path != save_path
    assert model_future.id in model_future.save_path
    assert os.path.exists(model_future.save_path)


def test_save_with_id_and_model_name(trainer_type_cfg, save_path):
    """Test that saving with the training id and model name
    correctly injects the ID and name in the save path
    """
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        save_path=save_path,
        save_with_id=True,
        model_name="abc",
    )
    model_future.wait()
    assert model_future.save_path != save_path
    assert model_future.id in model_future.save_path
    assert "abc" in model_future.save_path
    assert os.path.exists(model_future.save_path)


def test_save_with_model_name(trainer_type_cfg, save_path):
    """Test that saving with the model name correctly
    injects the model name in the save path
    """
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        save_path=save_path,
        save_with_id=False,
        model_name="abc",
    )
    model_future.wait()
    assert model_future.save_path != save_path
    assert model_future.id not in model_future.save_path
    assert "abc" in model_future.save_path
    assert os.path.exists(model_future.save_path)


def test_cancel_clean_termination(trainer_type_cfg):
    """Test that cancelling an in-progress training successfully destroys the
    training when the training is run in a way that can be shut down cleanly
    """
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        sleep_time=1000,
    )
    assert model_future.get_info().status == TrainingStatus.RUNNING
    assert not model_future.get_info().status.is_terminal

    # Cancel the future
    model_future.cancel()
    assert model_future.get_info().status == TrainingStatus.CANCELED
    assert model_future.get_info().status.is_terminal
    model_future.wait()


def test_cancel_without_waiting(trainer_type_cfg):
    """Test that cancelling an in-progress training that uses a long sleep (and
    can't be easily destroyed in a thread) still reports CANCELED as the status
    before the training has fully terminated.
    """
    trainer = local_trainer(**trainer_type_cfg)
    model_future = trainer.train(
        SampleModule,
        training_data=DataStream.from_iterable([]),
        sleep_time=0.5,
        sleep_increment=0.5,
    )
    assert model_future.get_info().status == TrainingStatus.RUNNING
    assert not model_future.get_info().status.is_terminal

    # Cancel the future and make sure it reports canceled, even though the
    # function is still sleeping
    model_future.cancel()
    assert model_future.get_info().status == TrainingStatus.CANCELED
    assert model_future.get_info().status.is_terminal


def test_no_retention_time(trainer_type_cfg):
    """Test that constructing with no configured retention period keeps futures
    forever and doesn't cause errors
    """
    trainer = local_trainer(retention_duration=None, **trainer_type_cfg)
    model_future = trainer.train(SampleModule, DataStream.from_iterable([]))
    model_future.wait()
    retrieved_future = trainer.get_model_future(model_future.id)
    assert retrieved_future is model_future


def test_purge_retention_time(trainer_type_cfg):
    """Test that purging old models works as expected"""
    trainer = local_trainer(retention_duration="1d10s", **trainer_type_cfg)
    model_future = trainer.train(SampleModule, DataStream.from_iterable([]))
    model_future.wait()
    retrieved_future = trainer.get_model_future(model_future.id)
    assert retrieved_future is model_future
    model_future._completion_time = model_future._completion_time - timedelta(days=2)
    with pytest.raises(CaikitCoreException):
        trainer.get_model_future(model_future.id)


@pytest.mark.parametrize(
    "test_params",
    [
        ("1d", timedelta(days=1)),
        ("0.01s", timedelta(seconds=0.01)),
        ("1d12h3m0.2s", timedelta(days=1, hours=12, minutes=3, seconds=0.2)),
    ],
)
def test_retention_duration_parse(test_params):
    """Make sure the regex for the duration can parse all expected durations"""
    trainer = local_trainer(retention_duration=test_params[0])
    assert trainer._retention_duration == test_params[1]


def test_get_into_return_error(trainer_type_cfg):
    """Test that failed training returns error properly"""
    trainer = local_trainer(**trainer_type_cfg)

    model_future = trainer.train(
        SampleModule,
        DataStream.from_iterable([]),
        batch_size=SampleModule.POISON_PILL_BATCH_SIZE,
    )
    # assert model_future.get_info().status == TrainingStatus.RUNNING
    # assert not model_future.get_info().status.is_terminal

    # Let the training proceed and wait for it to complete
    model_future.wait()
    assert model_future.get_info().status == TrainingStatus.ERRORED
    assert model_future.get_info().status.is_terminal
    assert isinstance(model_future.get_info().errors, list)
    assert isinstance(model_future.get_info().errors[0], ValueError)
    assert str(model_future.get_info().errors[0]) == "Batch size of 999 is not allowed!"



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Local saver implementation for saving modules to disk.
Contains recursive functions for loading modules saved inside modules.
"""

# Standard
from importlib import metadata
import datetime
import os
import shutil
import uuid

# First Party
import alog

# Local
from ..exceptions import error_handler
from ..modules.config import ModuleConfig
from ..toolkit import ObjectSerializer
from ..toolkit.wip_decorator import TempDisableWIP
from .base import ModuleBase
from .loader import ModuleLoader
from caikit.config import get_config

log = alog.use_channel("MODULE_SAVE")
error = error_handler.get(log)


class ModuleSaver:
    """A module saver that provides common functionality used for saving modules and also a context
    manager that cleans up gracefully in case an error is encountered during the save process.
    """

    SAVED_KEY_NAME = "saved"
    CREATED_KEY_NAME = "created"
    TRACKING_KEY_NAME = "tracking_id"
    MODULE_VERSION_KEY_NAME = "version"
    MODULE_ID_KEY_NAME = "module_id"
    MODULE_CLASS_KEY_NAME = "module_class"

    def __init__(self, module: ModuleBase, model_path):
        """Construct a new module saver.

        Args:
            module (caikit.core.module.Module): The instance of the module to be
                saved.
            model_path (str): The absolute path to the directory where the model
                will be saved.  If this directory does not exist, it will be
                created.
        """
        self.model_path = os.path.normpath(model_path)

        # Get possibly nested caikit library path
        module_path = module.__module__
        lib_name_generator = (
            k
            for k, v in get_config().libraries.items()
            if module_path.startswith(v.module_path)
        )
        try:
            self.library_name = next(lib_name_generator)
        except StopIteration:
            # This assumes no nested module path by default
            self.library_name = module_path.split(".")[0]  # tests

        try:
            self.library_version = metadata.version(self.library_name)
        except metadata.PackageNotFoundError:
            log.debug("<COR25991305D>", "No library version found")
            if (
                self.library_name in get_config().libraries
                and "version" in get_config().libraries[self.library_name]
            ):
                self.library_version = get_config().libraries[self.library_name].version
            else:
                self.library_version = "0.0.0"

        self.config = {
            self.library_name + "_version": self.library_version,
            self.CREATED_KEY_NAME: str(datetime.datetime.now()),
            self.SAVED_KEY_NAME: str(datetime.datetime.now()),
            "name": module.MODULE_NAME,
            self.TRACKING_KEY_NAME: str(uuid.uuid4()),
            self.MODULE_ID_KEY_NAME: module.MODULE_ID,
            self.MODULE_CLASS_KEY_NAME: module.MODULE_CLASS,
            self.MODULE_VERSION_KEY_NAME: module.MODULE_VERSION,
        }

        # Temp disable wip for following invocation to not log warnings for downstream
        # usage of ModuleSaver
        with TempDisableWIP():
            # Get metadata back about this module and add it to the config
            stored_config = module.metadata
        # Sanitize some things off of the config:
        # Remove the old `saved` timestamp:
        stored_config.pop(self.SAVED_KEY_NAME, None)
        # Remove any reserved keys, these will be set by the `ModuleConfig` class
        for key in ModuleConfig.reserved_keys:
            if key in stored_config:
                stored_config.pop(key)

        self.config.update(stored_config)

    def add_dir(self, relative_path, base_relative_path=""):
        """Create a directory inside the `model_path` for this saver.

        Args:
            relative_path (str): A path relative to this saver's `model_path`
                denoting the directory to create.
            base_relative_path (str): A path, relative to this saver's
                `model_path`, in which `relative_path` will be created.

        Returns:
            str, str: A tuple containing both the `relative_path` and
                `absolute_path` to the directory created.

        Examples:
            >>> with ModelSaver('/path/to/model') as saver:
            >>>     rel_path, abs_path = saver.add_dir('word_embeddings', 'model_data')
            >>> print(rel_path)
            model_data/word_embeddings
            >>> print(abs_path)
            /path/to/model/model_data/word_embeddings
        """
        base_relative_path = os.path.normpath(base_relative_path)
        relative_path = os.path.normpath(relative_path)

        relative_path = os.path.join(base_relative_path, relative_path)
        absolute_path = os.path.join(self.model_path, relative_path)

        os.makedirs(absolute_path, exist_ok=True)

        return relative_path, absolute_path

    def copy_file(self, file_path, relative_path=""):
        """Copy an external file into a subdirectory of the `model_path` for this saver.

        Args:
            file_path (str): Absolute path to the external file to copy.
            relative_path (str): The relative path inside of `model_path` where
                the file will be copied to. If set to the empty string (default)
                then the file will be placed directly in the `model_path`
                directory.

        Returns:
            str, str: A tuple containing both the `relative_path` and
                `absolute_path` to the copied file.
        """
        file_path = os.path.normpath(file_path)

        if not os.path.isfile(file_path):
            error(
                "<COR80954473E>",
                FileNotFoundError(
                    "Attempted to add `{}` but is not a regular file.".format(file_path)
                ),
            )

        filename = os.path.basename(os.path.normpath(file_path))

        relative_path, absolute_path = self.add_dir(relative_path)

        relative_file_path = os.path.join(relative_path, filename)
        absolute_file_path = os.path.join(absolute_path, filename)

        shutil.copyfile(file_path, absolute_file_path)

        return relative_file_path, absolute_file_path

    def save_object(self, obj, filename, serializer, relative_path=""):
        """Save a Python object using the provided ObjectSerializer.

        Args:
            obj (any): The Python object to save
            filename (str): The filename to use for the saved object
            serializer (ObjectSerializer): An ObjectSerializer instance (e.g.,
                YAMLSerializer) that should be used to serialize the object
            relative_path (str): The relative path inside of `model_path` where
                the object will be saved
        """
        if not issubclass(serializer.__class__, ObjectSerializer):
            error(
                "<COR85655282E>",
                TypeError(
                    "`{}` does not extend `ObjectSerializer`".format(
                        serializer.__class__.__name__
                    )
                ),
            )

        relative_path, absolute_path = self.add_dir(relative_path)

        # Normalize any '././' structure that may come from relative paths
        relative_file_path = os.path.normpath(os.path.join(relative_path, filename))
        absolute_file_path = os.path.normpath(os.path.join(absolute_path, filename))

        serializer.serialize(obj, absolute_file_path)

        return relative_file_path, absolute_file_path

    def update_config(self, additional_config):
        """Add items to this saver's config dictionary.

        Args:
            additional_config (dict): A dictionary of config options to add the
                this saver's configuration.

        Notes:
            The behavior of this method matches `dict.update` and is equivalent to calling
            `saver.config.update`.  The `saver.config` dictionary may be accessed directly for
            more sophisticated manipulation of the configuration.
        """
        self.config.update(additional_config)

    def save_module(self, module, relative_path, **kwargs):
        """Save a CaikitCore module within a workflow artifact and add a reference to the config.

        Args:
            module (caikit.core.ModuleBase): The CaikitCore module to save as
                part of this workflow
            relative_path (str): The relative path inside of `model_path` where
                the module will be saved
            **kwargs:  dict
                key-value pair of parameters to be passed to module.save
        """

        if not issubclass(module.__class__, ModuleBase):
            error(
                "<COR30664151E>",
                TypeError(
                    "`{}` does not extend `ModuleBase`".format(
                        module.__class__.__name__
                    )
                ),
            )

        rel_path, abs_path = self.add_dir(relative_path)
        # Save this module at the specified location
        module.save(abs_path, **kwargs)
        self.config.setdefault(ModuleLoader.MODULE_PATHS_KEY, {}).update(
            {relative_path: rel_path}
        )
        return rel_path, abs_path

    def save_module_list(self, modules, config_key, **kwargs):
        """Save a list of CaikitCore modules within a workflow artifact and add a reference to the
        config.

        Args:
            modules (dict{str -> caikit.core.ModuleBase}): A dict with module
                relative path as key and a CaikitCore module as value to save as
                part of this workflow
            config_key (str): The config key inside of `model_path` where the
                modules' relative path with be referenced
            **kwargs:  dict
                key-value pair of parameters to be passed to module.save


        Returns:
            list_of_rel_path: list(str)
                List of relative paths where the modules are saved
            list_of_abs_path: list(str)
                List of absolute paths where the modules are saved
        """
        # validate type of input parameters
        error.type_check("<COR44644420E>", dict, modules=modules)
        error.type_check("<COR54316176E>", str, config_key=config_key)

        list_of_rel_path = []
        list_of_abs_path = []

        # iterate through the dict and serialize the modules in its corresponding paths
        for relative_path, module in modules.items():
            if not issubclass(module.__class__, ModuleBase):
                error(
                    "<COR67834055E>",
                    TypeError(
                        "`{}` does not extend `ModuleBase`".format(
                            module.__class__.__name__
                        )
                    ),
                )
            error.type_check("<COR48984754E>", str, relative_path=relative_path)

            rel_path, abs_path = self.add_dir(relative_path)

            # Save this module at the specified location
            module.save(abs_path, **kwargs)

            # append relative and absolute path to a list that will be returned
            list_of_rel_path.append(rel_path)
            list_of_abs_path.append(abs_path)

        # update the config with config key and list of relative path
        self.config.setdefault(ModuleLoader.MODULE_PATHS_KEY, {}).update(
            {config_key: list_of_rel_path}
        )
        return list_of_rel_path, list_of_abs_path

    def __enter__(self):
        """Enter the module saver context.  This creates the `model_path` directory.  If this
        context successfully exits, then the model configuration and all files it contains will
        be written and saved to disk inside the `model_path` directory.  If any uncaught exceptions
        are thrown inside this context, then `model_path` will be removed.
        """
        os.makedirs(self.model_path, exist_ok=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the module saver context. If this context successfully exits, then the model
        configuration and all files it contains will be written and saved to disk inside the
        `model_path` directory.  If any uncaught exceptions are thrown inside this context, then
        `model_path` will be removed.
        """
        if exc_type is not None:
            shutil.rmtree(self.model_path, ignore_errors=True)
            return

        ModuleConfig(self.config).save(self.model_path)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Third Party
import pytest

# Local
from caikit.core.augmentors import AugmentorBase, MergedAugmentor, schemes

# Unit Test Infrastructure
from tests.base import TestCaseBase


class StubStringAugmentor(AugmentorBase):
    augmentor_type = str
    # Indicates that this class should not be picked up by augmentor discovery utilities.
    is_test_augmentor = True

    def __init__(self, identifier):
        super().__init__(random_seed=1001)
        self.id = identifier

    def _augment(self, obj):
        return "{} {}".format(obj, self.id)


class StubIntegerAugmentor(AugmentorBase):
    augmentor_type = int
    # Indicates that this class should not be picked up by augmentor discovery utilities.
    is_test_augmentor = True

    def __init__(self, identifier):
        super().__init__(random_seed=1001)
        self.id = identifier

    def _augment(self, obj):
        return obj


class StubNoneProducingAugmentor(AugmentorBase):
    """Test class augmentor schemes / merge augmentors."""

    augmentor_type = int
    # Indicates that this class should not be picked up by augmentor discovery utilities.
    is_test_augmentor = True

    def __init__(self):
        super().__init__(random_seed=1001, produces_none=True)


class TestAugmentorSchemes(TestCaseBase):
    @classmethod
    def setUpClass(cls):
        cls.augmentors = [StubStringAugmentor("first"), StubStringAugmentor("second")]

    ###############################################################################################
    #                        Happy path cases for all Augmentor Schemes                           #
    ###############################################################################################
    # Note: For the tests below, since we call _execute(), not execute(), preserve order
    # value does not matter in these tests. execute() is a thin wrapper around ._execute()
    # that has an alternate path for random shuffling.
    def test_always_selection_scheme(self):
        """Ensure that an "always" selection scheme applies all augmentors."""
        always_scheme = schemes.AlwaysSelectionScheme(True, self.augmentors)
        self.assertEqual(always_scheme._execute("text"), "text first second")

    def test_multi_selection_scheme(self):
        """Random state independent test cases for random multi selection scheme."""
        always_multi_scheme = schemes.RandomMultiSelectionScheme(
            True, [1, 1], self.augmentors
        )
        never_multi_scheme = schemes.RandomMultiSelectionScheme(
            True, [0, 0], self.augmentors
        )
        self.assertEqual(always_multi_scheme._execute("text"), "text first second")
        self.assertEqual(never_multi_scheme._execute("text"), "text")

    def test_single_selection_scheme(self):
        """Random state independent test cases for random single selection scheme."""
        first_scheme = schemes.RandomSingleSelectionScheme([1, 0], self.augmentors)
        second_scheme = schemes.RandomSingleSelectionScheme([0, 1], self.augmentors)
        self.assertEqual(first_scheme._execute("text"), "text first")
        self.assertEqual(second_scheme._execute("text"), "text second")

    ###############################################################################################
    #                   Input validation for Augmentor Scheme initialization                      #
    ###############################################################################################
    # Base class
    def test_base_checks_preserve_order_type(self):
        """Ensure that preserve order needs to be a boolean value."""
        self.assertRaises(TypeError, schemes.SchemeBase, None, self.augmentors, 1001)

    def test_base_validates_augmentor_types(self):
        """Ensure SchemeBase type checks augmentor list/tuple properly."""
        self.assertRaises(TypeError, schemes.SchemeBase, True, "Foobar", 1001)
        self.assertRaises(TypeError, schemes.SchemeBase, True, ["Foobar"], 1001)

    def test_base_requires_augmentors(self):
        """Ensure SchemeBase requires at least one augmentor."""
        self.assertRaises(ValueError, schemes.SchemeBase, True, [], 1001)

    # Random Multi Selection Scheme
    def test_multi_selection_prob_types(self):
        """Ensure that selection probability types are handled correctly."""
        self.assertRaises(
            TypeError,
            schemes.RandomMultiSelectionScheme,
            True,
            ["a", "b"],
            self.augmentors,
            1001,
        )
        self.assertRaises(
            TypeError,
            schemes.RandomMultiSelectionScheme,
            True,
            "foo",
            self.augmentors,
            1001,
        )

    def test_multi_selection_prob_length_alignment(self):
        """Ensure that the number of probability values needs to match the number of augmentors."""
        self.assertRaises(
            ValueError,
            schemes.RandomMultiSelectionScheme,
            True,
            [1],
            self.augmentors,
            1001,
        )
        self.assertRaises(
            ValueError,
            schemes.RandomMultiSelectionScheme,
            True,
            [1, 2, 3],
            self.augmentors,
            1001,
        )

    def test_multi_selection_prob_range(self):
        """Ensure that selection probability ranges are properly validated."""
        self.assertRaises(
            ValueError,
            schemes.RandomMultiSelectionScheme,
            True,
            [-1, 2],
            self.augmentors,
            1001,
        )

    # Random Single Selection Scheme
    def test_single_selection_prob_types(self):
        """Ensure that selection probability types are handled correctly."""
        self.assertRaises(
            TypeError,
            schemes.RandomSingleSelectionScheme,
            ["a", "b"],
            self.augmentors,
            1001,
        )
        self.assertRaises(
            TypeError, schemes.RandomSingleSelectionScheme, "foo", self.augmentors, 1001
        )

    def test_single_selection_prob_length_alignment(self):
        """Ensure that the number of probability values needs to match the number of augmentors."""
        self.assertRaises(
            ValueError, schemes.RandomSingleSelectionScheme, [1], self.augmentors, 1001
        )
        self.assertRaises(
            ValueError,
            schemes.RandomSingleSelectionScheme,
            [1, 2, 3],
            self.augmentors,
            1001,
        )

    def test_single_selection_prob_sum_must_be_one(self):
        """Ensure that the selection probability sum of a select one problem must be one."""
        self.assertRaises(
            ValueError,
            schemes.RandomSingleSelectionScheme,
            [0.2, 0.4],
            self.augmentors,
            1001,
        )

    def test_single_selection_prob_range(self):
        """Ensure that selection probability ranges are properly validated."""
        self.assertRaises(
            ValueError,
            schemes.RandomSingleSelectionScheme,
            [-1, 2],
            self.augmentors,
            1001,
        )


class TestMergedAugmentor(TestCaseBase):
    def test_incompatible_augmentor_type_initialization(self):
        """Ensure that we raise if incompatible augmentors are provided."""
        str_aug = StubStringAugmentor("text")
        int_aug = StubIntegerAugmentor(13)
        scheme = schemes.RandomSingleSelectionScheme([1, 0], [str_aug, int_aug])
        self.assertRaises(ValueError, MergedAugmentor, scheme)

    def test_compatible_augmentor_type_initialization(self):
        """Ensure that we can build a Merged Augmentor with a well-formed merging scheme."""
        str_aug = StubStringAugmentor("text")
        scheme = schemes.RandomSingleSelectionScheme([1, 0], [str_aug, str_aug])
        aug = MergedAugmentor(scheme)
        self.assertEqual(aug.augmentor_type, str)

    def test_scheme_type_validation(self):
        """Ensure that MergedAugmentor raises if a bad arg type is provided for scheme."""
        self.assertRaises(TypeError, MergedAugmentor, 13)

    def test_merge_with_produces_none_fails(self):
        """Ensure that we cannot merge augmentors if any of them can produce <None>."""
        int_aug = StubIntegerAugmentor(13)
        none_int_aug = StubNoneProducingAugmentor()
        scheme = schemes.RandomSingleSelectionScheme([1, 0], [int_aug, none_int_aug])
        self.assertRaises(ValueError, MergedAugmentor, scheme)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the reversible hasher
"""

# Standard
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import random

# Local
from caikit.core.toolkit.reversible_hasher import ReversibleHasher

## Helpers #####################################################################


def do_hash(x):
    hashed = ReversibleHasher.hash(x)
    reversed_ = ReversibleHasher.reverse_hash(hashed)
    return (hashed, reversed_)


## Tests #######################################################################


def test_reversible_hasher_single_process():
    """Make sure that hashing the same string multiple times in sequence is
    repeatable and reversible
    """
    sample_string = "thisisatest1234-of_stuff"
    hashed1 = ReversibleHasher.hash(sample_string)
    hashed2 = ReversibleHasher.hash(sample_string)
    reversed1 = ReversibleHasher.reverse_hash(hashed1)
    reversed2 = ReversibleHasher.reverse_hash(hashed2)
    assert hashed1 == hashed2
    assert reversed1 == reversed2 == sample_string


def test_reversible_hasher_multi_thread():
    """Make sure that hashing the same string multiple times in different
    threads is repeatable and reversible
    """
    sample_string = "ANoTHe-r_t327"
    executor = ThreadPoolExecutor()
    future1 = executor.submit(do_hash, x=sample_string)
    future2 = executor.submit(do_hash, x=sample_string)
    (hashed1, reversed1) = future1.result()
    (hashed2, reversed2) = future2.result()
    assert hashed1 == hashed2
    assert reversed1 == reversed2 == sample_string


def test_reversible_hasher_multi_process():
    """Make sure that hashing the same string multiple times in different
    processes is repeatable and reversible
    """
    sample_string = "y3t_ANoTHe-r_tES T"
    executor = ProcessPoolExecutor()
    future1 = executor.submit(do_hash, x=sample_string)
    future2 = executor.submit(do_hash, x=sample_string)
    (hashed1, reversed1) = future1.result()
    (hashed2, reversed2) = future2.result()
    assert hashed1 == hashed2
    assert reversed1 == reversed2 == sample_string


def test_reversible_hasher_random_still_random():
    """Make sure that the random state is not changed by hashing"""
    rand_state = random.getstate()
    do_hash("yetanother")
    assert random.getstate() == rand_state



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Base classes and functionality for all data structures.
"""
# Standard
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union
import base64
import datetime
import json

# Third Party
from google.protobuf import json_format
from google.protobuf.descriptor import Descriptor, FieldDescriptor, OneofDescriptor
from google.protobuf.internal import type_checkers as proto_type_checkers
from google.protobuf.message import Message as ProtoMessageType

# First Party
import alog

# Local
from ..exceptions import error_handler
from . import enums, json_dict, timestamp

# metaclass-generated field members cannot be detected by pylint
# pylint: disable=no-member
# pylint: disable=too-many-lines


log = alog.use_channel("DATAM")
error = error_handler.get(log)


class _DataBaseMetaClass(type):
    """Meta class for all structures in the data model."""

    # store a registry of all classes that use this metaclass, i.e.,
    # all classes that extend DataBase. This is used for constructing new
    # instances by name without having to introspect all modules in data_model.
    class_registry = {}

    # This sentinel value is used to determine whether a given attribute is
    # present on a class without doing `getattr` twice in the case where the
    # attribute does exist.
    _MISSING_ATTRIBUTE = "missing attribute"

    # Special attribute used to communicate that the proto fields are forward
    # declared and will be populated after the metaclass has completed
    # construction.
    _FWD_DECL_FIELDS = "__fwd_decl_fields__"

    # Special instance attributes that an instance of a class derived from
    # DataBase may have. These are added to __slots__.
    _BACKEND_ATTR = "_backend"
    _WHICH_ONEOF_ATTR = "_which_oneof"

    # When inferring which field in a oneof a given value should be used for
    # based on the python type, we need to check types in order with bool first,
    # ints next, then floats values that fit a "more flexible" type don't
    # accidentally get assigned to the wrong field. These are the lists of int
    # and bool type values in protobuf.
    _PROTO_TYPE_ORDER = [FieldDescriptor.TYPE_BOOL] + [
        val
        for name, val in vars(FieldDescriptor).items()
        if name.startswith("TYPE_") and "INT" in name
    ]

    def __new__(mcs, name, bases, attrs):
        """When constructing a new data model class, we set the 'fields' class variable from the
        protobufs descriptor and then set the '__slots__' magic class attribute to fields.  This
        provides two benefits: (a) performance is improved since the classes only need to know
        about these attributes (b) it helps to enforce that all member variables in these classes
        are described in the protobufs.

        Note:  If you want to add a variable for internal use that is not described in the
        protobufs, it can be named in the tuple class variable _private_slots and will
        automatically be added to __slots__.
        """

        # Protobufs fields can be divided into these categories, which are used
        # to automatically determine appropriate behavior in a number of methods
        attrs["full_name"] = name
        attrs["fields_enum_map"] = {}
        attrs["fields_enum_rev"] = {}
        attrs["_fields_oneofs_map"] = {}
        attrs["_fields_to_oneof"] = {}
        attrs["_fields_map"] = ()
        attrs["_fields_message"] = ()
        attrs["_fields_message_repeated"] = ()
        attrs["_fields_enum"] = ()
        attrs["_fields_enum_repeated"] = ()
        attrs["_fields_primitive"] = ()
        attrs["_fields_primitive_repeated"] = ()

        # Look for the set of fields either from a predefined protobuf class or
        # from a forward declaration from @dataobject
        fields = ()
        proto_class = None
        if name not in ["DataBase", "DataObjectBase"]:
            # Look for a precompiled proto class and if found, parse its
            # descriptor
            proto_class = attrs.get("_proto_class")
            if proto_class is not None:
                all_oneof_fields = [
                    field.name
                    for oneof in proto_class.DESCRIPTOR.oneofs
                    for field in oneof.fields
                ]
                fields = tuple(
                    (
                        field
                        for field in proto_class.DESCRIPTOR.fields_by_name
                        if field not in all_oneof_fields
                    )
                ) + tuple(proto_class.DESCRIPTOR.oneofs_by_name)

            # Otherwise, we need to get the fields from a "special" attribute
            else:
                fields = attrs.pop(mcs._FWD_DECL_FIELDS, None)
                log.debug4(
                    "Using dataclass forward declaration fields %s for %s", fields, name
                )
                error.value_check(
                    "<COR49310991E>",
                    fields is not None,
                    "No proto class found for {}",
                    name,
                )
        attrs["fields"] = fields
        attrs["_proto_class"] = proto_class

        # Look if any private slots are declared as class variables
        private_slots = attrs.setdefault("_private_slots", ())

        # Class slots are fields + private slots, this prevents other
        # member attributes from being set and also improves performance
        attrs["__slots__"] = tuple(
            [f"_{field}" for field in fields]
            + list(private_slots)
            + [mcs._BACKEND_ATTR, mcs._WHICH_ONEOF_ATTR]
        )

        # Create the instance of the type
        instance = super().__new__(mcs, name, bases, attrs)

        # If there's a valid proto class, perform proto descriptor parsing
        if proto_class is not None:
            mcs.parse_proto_descriptor(instance)

        # Return the constructed class instance
        return instance

    @classmethod
    def parse_proto_descriptor(mcs, cls):
        """Encapsulate the logic for parsing the protobuf descriptor here. This
        allows the parsing to be done as a post-process after metaclass
        initialization
        """

        # use the fully qualified protobuf name to avoid conflicts with
        # nested messages that have matching names
        cls.full_name = cls._proto_class.DESCRIPTOR.full_name

        # preserve old fields for _make_property_getter later
        old_fields = cls.fields

        # overwrite to only have proto-specific fields present
        cls.fields = tuple(cls._proto_class.DESCRIPTOR.fields_by_name)

        # map from all enum fields to their enum classes
        # note: enums are also primitives, these overlap
        cls.fields_enum_map = {
            field.name: getattr(enums, field.enum_type.name)
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.enum_type is not None
        }

        cls.fields_enum_rev = {
            field.name: getattr(enums, field.enum_type.name + "Rev")
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.enum_type is not None
        }

        # mapping of all oneofs and the fields that are part of them
        # NOTE: protobuf makes an interesting use of oneof to wrap types that
        #   should be explicitly optional. We don't want to consider these
        #   oneofs in the general oneof handling.

        # Sort the names of the fields in this map to ensure that ordering is
        # correct such that bool < int < float

        cls._fields_oneofs_map = {
            oneof_name: mcs._sorted_oneof_field_names(oneof)
            for oneof_name, oneof in cls._proto_class.DESCRIPTOR.oneofs_by_name.items()
            if len(oneof.fields) != 1 or oneof.name != f"_{oneof.fields[0].name}"
        }
        cls._fields_to_oneof = {
            field_name: oneof_name
            for (oneof_name, oneof_fields) in cls._fields_oneofs_map.items()
            for field_name in oneof_fields
        }

        # all repeated fields
        fields_repeated = tuple(
            field.name
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.label == field.LABEL_REPEATED
        )

        # all messages, repeated or not
        _fields_message_all = tuple(
            field.name
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.type == field.TYPE_MESSAGE
        )

        # all enums, repeated or not
        _fields_enum_all = tuple(
            field.name
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.enum_type is not None
        )

        # all fields of type map
        cls._fields_map = tuple(
            field.name
            for field in cls._proto_class.DESCRIPTOR.fields
            if field.message_type and field.message_type.GetOptions().map_entry
        )

        # all primitives, repeated or not
        _fields_primitive_all = (
            frozenset(cls.fields)
            .difference(cls._fields_map)
            .difference(_fields_message_all)
            .difference(_fields_enum_all)
        )

        # messages that are not repeated
        cls._fields_message = frozenset(_fields_message_all).difference(fields_repeated)

        # messages that are repeated
        cls._fields_message_repeated = frozenset(fields_repeated).intersection(
            _fields_message_all
        )

        # enums that are not repeated
        cls._fields_enum = frozenset(_fields_enum_all).difference(fields_repeated)

        # enums that are repeated
        cls._fields_enum_repeated = frozenset(_fields_enum_all).intersection(
            fields_repeated
        )

        # primitives that are not repeated
        cls._fields_primitive = frozenset(_fields_primitive_all).difference(
            fields_repeated
        )

        # primitives that are repeated
        cls._fields_primitive_repeated = frozenset(fields_repeated).intersection(
            _fields_primitive_all
        )

        # Update the global class and proto registries
        # NOTE: Explicitly not respecting metaclass inheritance so single
        #   registry shared for all
        _DataBaseMetaClass.class_registry[cls.full_name] = cls

        # Add properties that use the underlying backend. Also add fields that
        # existed in old_fields for supporting oneofs
        # see https://github.com/caikit/caikit/pull/107 for details
        for field in set(cls.fields + tuple(old_fields)):
            # If the field is the name of a field within a oneof and it was not
            # in the old fields, the data is held under the oneof's name if this
            # is the set value for the oneof
            if oneof_name := cls._fields_to_oneof.get(field):
                setattr(cls, field, mcs._make_property_getter(field, oneof_name))

            # If the field is a plain field or the name of a oneof, it will be
            # accessed directly
            else:
                setattr(cls, field, mcs._make_property_getter(field))

        # If there is not already an __init__ function defined, make one
        current_init = cls.__init__
        if current_init is None or current_init is DataBase.__init__:
            setattr(cls, "__init__", mcs._make_init(cls.fields))

    @classmethod
    def _make_property_getter(mcs, field, oneof_name=None):
        """This helper creates an @property attribute getter for the given field

        NOTE: This needs to live as a standalone function in order for the given
            field name to be properly bound to the closure for the attrs
        """
        private_name = f"_{field}" if oneof_name is None else oneof_name

        def _property_getter(self):
            # Check to see if the private name is defined and just return it if
            # it is
            current = getattr(self, private_name, mcs._MISSING_ATTRIBUTE)
            if current is not mcs._MISSING_ATTRIBUTE:
                return current

            # If not currently set, delegate to the backend
            backend = self.backend
            if backend is None:
                error(
                    "<COR66616239E>",
                    AttributeError(
                        f"{type(self)} missing attribute {field} and no backend set"
                    ),
                )
            attr_val = backend.get_attribute(self.__class__, field)
            if isinstance(attr_val, self.__class__.OneofFieldVal):
                log.debug2("Got a OneofFieldVal from the backend")
                assert field in self.__class__._fields_oneofs_map
                self._get_which_oneof_dict()[field] = attr_val.which_oneof
                attr_val = attr_val.val

            # If the backend says that this attribute should be cached, set it
            # as an attribute on the class
            if backend.cache_attribute(field, attr_val):
                setattr(self, field, attr_val)

            # Return the value found by the backend
            return attr_val

        # If this is a oneof, add an extra layer of wrapping to check
        # which_oneof before returning a valid result
        if oneof_name:

            def _oneof_property_getter(self):
                if self.which_oneof(oneof_name) == field:
                    return _property_getter(self)

            return property(_oneof_property_getter)

        return property(_property_getter)

    @staticmethod
    def _make_init(fields):
        """This helper creates an __init__ function for a class which has the
        arguments for all the fields and just sets them as instance attributes.
        """

        # Format and preserve docstring
        docstring = """Construct with arguments for each field on the object
            Args:
                {}
        """.format(
            "\n    ".join(fields)
        )

        def __init__(self, *args, **kwargs):
            num_args = len(args)
            num_kwargs = len(kwargs)
            num_fields = len(fields)
            used_fields = []

            # If the proto has oneofs, set up which_oneof
            which_oneof = {}
            cls = self.__class__
            if cls._fields_oneofs_map:
                setattr(self, _DataBaseMetaClass._WHICH_ONEOF_ATTR, which_oneof)

            if num_args + num_kwargs > num_fields:
                error(
                    "<COR71444420E>",
                    TypeError(f"Too many arguments given. Args are: {fields}"),
                )

            if num_args > 0:  # Do a quick check for performance reason
                for i, field_val in enumerate(args):
                    field_name = fields[i]
                    setattr(self, field_name, field_val)
                    used_fields.append(field_name)

            if num_kwargs > 0:  # Do a quick check for performance reason
                for field_name, field_val in kwargs.items():
                    # If this is a oneof field, alias to the oneof name
                    if oneof_name := cls._fields_to_oneof.get(field_name):
                        which_oneof[oneof_name] = field_name
                        field_name = oneof_name

                    if (
                        field_name not in fields
                        and field_name not in cls._fields_oneofs_map
                    ):
                        error(
                            "<COR71444421E>", TypeError(f"Unknown field {field_name}")
                        )
                    elif field_name in used_fields:
                        error(
                            "<COR71444422E>",
                            TypeError(f"Got multiple values for field {field_name}"),
                        )
                    setattr(self, field_name, field_val)
                    used_fields.append(field_name)

            # Default all unspecified fields to None
            if num_fields > 0:  # Do a quick check for performance reason
                for field_name in fields:
                    if (
                        field_name not in used_fields
                        and field_name not in cls._fields_to_oneof
                    ):
                        setattr(self, field_name, None)

        # Set docstring to the method explicitly
        setattr(__init__, "__doc__", docstring)
        return __init__

    @classmethod
    def _sorted_oneof_field_names(mcs, oneof: OneofDescriptor) -> List[str]:
        """Helper to get the list of oneof fields while ensuring field names are
        sorted such that bool < int < float. This ensures that when iterating
        fields for which_oneof inference, lower-precedence types take
        precedence.
        """
        return [
            field.name
            for field in sorted(
                oneof.fields,
                key=lambda fld: mcs._PROTO_TYPE_ORDER.index(fld.type)
                if fld.type in mcs._PROTO_TYPE_ORDER
                else len(mcs._PROTO_TYPE_ORDER),
            )
        ]


class DataBase(metaclass=_DataBaseMetaClass):
    """Base class for all structures in the data model.

    Notes:
        All leaves in the hierarchy of derived classes should have a corresponding protobufs class
        defined in the interface definitions.  If not, an exception will be thrown at runtime.
    """

    @dataclass
    class OneofFieldVal:
        """Helper struct that backends can use to return information about
        values in oneofs along with which of the oneofs is currently valid
        """

        val: Any
        which_oneof: str

    def __setattr__(self, name, val):
        """Handle attribute setting for oneofs and named fields with delegation
        to backends as needed
        """
        # If setting a oneof directly, remove any oneof information
        cls = self.__class__
        if name in cls._fields_oneofs_map:
            self._get_which_oneof_dict().pop(name, None)

        # If this is the name of a oneof field, set the oneof itself
        if oneof_name := cls._fields_to_oneof.get(name):
            self._get_which_oneof_dict()[oneof_name] = name
            name = oneof_name

        # If attempting to set one of the named fields or a oneof, instead set
        # the private version of the attribute.
        if name in cls.fields or name in cls._fields_oneofs_map:
            super().__setattr__(f"_{name}", val)
        else:
            super().__setattr__(name, val)

    @classmethod
    def get_proto_class(cls) -> Type[ProtoMessageType]:
        return cls._proto_class

    @classmethod
    def get_field_message_type(cls, field_name: str) -> Optional[Type["DataBase"]]:
        """Get the data model class for the given field if the field is a
        message or a repeated message

        Args:
            field_name (str): Field name to check (AttributeError raised if name
                is invalid)

        Returns:
            data_model_type:  Type[DataBase]
                The data model class type for the given field
        """
        if field_name not in cls.fields:
            raise AttributeError(f"Invalid field {field_name}")
        if (
            field_name in cls._fields_message
            or field_name in cls._fields_message_repeated
        ):
            return cls.get_class_for_proto(
                cls.get_proto_class().DESCRIPTOR.fields_by_name[field_name].message_type
            )
        return None

    @classmethod
    def from_backend(cls, backend):
        instance = cls.__new__(cls)
        setattr(instance, _DataBaseMetaClass._BACKEND_ATTR, backend)
        return instance

    @property
    def backend(self) -> Optional["DataModelBackendBase"]:
        return getattr(self, _DataBaseMetaClass._BACKEND_ATTR, None)

    def which_oneof(self, oneof_name: str) -> Optional[str]:
        """Get the name of the oneof field set for the given oneof or None if no
        field is set
        """
        # If the internal dict is already set, use that information
        which_oneof = self._get_which_oneof_dict()
        if current_val := which_oneof.get(oneof_name):
            return current_val

        # Get the current value for the oneof and introspect which field its
        # type matches
        oneof_val = getattr(self, oneof_name)

        # Re-check in case the getattr pulled a OneofFieldVal that populated the
        # which_oneof dict with knowledge from the backend
        if current_val := which_oneof.get(oneof_name):
            return current_val

        # Try to figure out the field based on the type
        which_field = self._infer_which_oneof(oneof_name, oneof_val)
        if which_field is not None:
            which_oneof[oneof_name] = which_field
        return which_field

    @classmethod
    def _infer_which_oneof(cls, oneof_name: str, oneof_val: Any) -> Optional[str]:
        """Check each candidate field within the oneof to see if it's a type
        match

        NOTE: In the case where fields within a oneof have the same type, the
          first field whose type matches will be used!
        """
        # NOTE: The list of field names are guaranteed to be sorted so that
        #   bool < int < float
        for field_name in cls._fields_oneofs_map.get(oneof_name, []):
            if cls._is_valid_type_for_field(field_name, oneof_val):
                return field_name

    def _get_which_oneof_dict(self) -> Dict[str, str]:
        which_oneof = getattr(self, _DataBaseMetaClass._WHICH_ONEOF_ATTR, None)
        if which_oneof is None:
            super().__setattr__(_DataBaseMetaClass._WHICH_ONEOF_ATTR, {})
            which_oneof = getattr(self, _DataBaseMetaClass._WHICH_ONEOF_ATTR)
        return which_oneof

    @classmethod
    def _is_valid_type_for_field(cls, field_name: str, val: Any) -> bool:
        """Check whether the given value is valid for the given field"""
        # pylint: disable=too-many-return-statements
        field_descriptor = cls._proto_class.DESCRIPTOR.fields_by_name[field_name]

        if val is None:
            return False

        # If val is a list, this maybe a union of list field
        if isinstance(val, list) and field_name in cls._fields_to_oneof:
            if len(val) == 0:
                log.info("Assuming the type is valid since list is empty")
                return True

            val_list_type = type(val[0]).__name__
            return (
                field_descriptor.message_type
                and f"{val_list_type}"
                in field_descriptor.message_type.full_name.lower()
            )

        # If it's a data object or an enum and the descriptors match, it's a
        # good type
        if (
            isinstance(val, DataBase)
            and field_descriptor.message_type == val.get_proto_class().DESCRIPTOR
        ) or (
            isinstance(val, Enum)
            and field_descriptor.enum_type == val.get_proto_class().DESCRIPTOR
        ):
            return True

        # If it's a data object or an enum and the descriptors don't match, it's
        # a bad type
        if field_descriptor.type in [
            field_descriptor.TYPE_MESSAGE,
            field_descriptor.TYPE_ENUM,
        ]:
            return False

        # If the field is a bool field, only accept python bools. Proto is ok to
        # accept ints, but we are stricter than that.
        if field_descriptor.type == field_descriptor.TYPE_BOOL:
            return isinstance(val, bool)

        # Proto doesn't allow non utf-8 bytes fields; however, python does.
        if field_descriptor.type == field_descriptor.TYPE_BYTES:
            return isinstance(val, bytes)

        # If it's a primitive, use protobuf type checkers
        checker = proto_type_checkers.GetTypeChecker(field_descriptor)
        try:
            checker.CheckValue(val)
            return True
        except TypeError:
            pass
        return False

    @classmethod
    def from_binary_buffer(cls, buf):
        """Builds the data model object out of the binary string

        Args:
            buf: The binary buffer containing a serialized protobufs message
        Returns:
            A data model object instantiated from the protobufs message deserialized out of `buf`
        """
        proto_message = cls.get_proto_class()()
        proto_message.ParseFromString(buf)

        return cls.from_proto(proto_message)

    @classmethod
    def from_proto(cls, proto):
        """Build a DataBase from protobufs.

        Args:
            proto: A protocol buffer to serialize from.
        Returns:
            protobufs: A DataBase object.
        """
        error.type_check("<COR45207671E>", ProtoMessageType, proto=proto)
        if cls._proto_class.DESCRIPTOR.name != proto.DESCRIPTOR.name:
            error(
                "<COR71783894E>",
                ValueError(
                    "class name `{}` does not match protobufs name `{}`".format(
                        cls._proto_class.DESCRIPTOR.name, proto.DESCRIPTOR.name
                    )
                ),
            )

        kwargs = {}
        for field in cls.fields:
            try:
                proto_attr = getattr(proto, field)
            except AttributeError:
                error(
                    "<COR71783905E>",
                    AttributeError(
                        "protobufs `{}` does not have field `{}`".format(
                            proto.DESCRIPTOR.name, field
                        )
                    ),
                )

            if field in cls._fields_primitive or field in cls._fields_enum:
                # special case for oneofs
                if field not in cls._fields_to_oneof or proto.HasField(field):
                    kwargs[field] = proto_attr
            elif (
                field in cls._fields_primitive_repeated
                or field in cls._fields_enum_repeated
            ):
                kwargs[field] = list(proto_attr)

            elif field in cls._fields_map:
                kwargs[field] = {}
                for key, value in proto_attr.items():
                    # Similar to filling; if our value is a non-primitive, i.e., a message,
                    # we need to look up the data model class attached to it.
                    if hasattr(value, "DESCRIPTOR"):
                        contained_class = cls.get_class_for_proto(value)
                        kwargs[field][key] = contained_class.from_proto(value)
                    # If it's not a message, the value can be left alone, i.e., it's a primitive
                    else:
                        kwargs[field][key] = value

            elif field in cls._fields_message:
                if proto.HasField(field):
                    if (
                        proto_attr.DESCRIPTOR.full_name
                        == json_dict.STRUCT_PROTOBUF_NAME
                    ):
                        kwargs[field] = json_dict.struct_to_dict(proto_attr)
                    elif (
                        proto_attr.DESCRIPTOR.full_name
                        == timestamp.TIMESTAMP_PROTOBUF_NAME
                    ):
                        kwargs[field] = timestamp.proto_to_datetime(proto_attr)
                    elif proto_attr.DESCRIPTOR.full_name.endswith("Sequence"):
                        oneof = cls._fields_to_oneof[field]
                        contained_class = cls.get_class_for_proto(proto_attr)
                        contained_obj = contained_class.from_proto(proto_attr)
                        kwargs[oneof] = getattr(contained_obj, "values")
                    else:
                        contained_class = cls.get_class_for_proto(proto_attr)
                        contained_obj = contained_class.from_proto(proto_attr)
                        kwargs[field] = contained_obj

            elif field in cls._fields_message_repeated:
                elements = []
                contained_class = None
                for item in proto_attr:
                    if item.DESCRIPTOR.full_name == json_dict.STRUCT_PROTOBUF_NAME:
                        elements.append(json_dict.struct_to_dict(item))
                    elif item.DESCRIPTOR.full_name == timestamp.TIMESTAMP_PROTOBUF_NAME:
                        elements.append(timestamp.proto_to_datetime(item))
                    else:
                        if contained_class is None:
                            contained_class = cls.get_class_for_proto(item)
                        elements.append(contained_class.from_proto(item))
                kwargs[field] = elements

            else:
                error(
                    "<COR71783815E>",
                    AttributeError(
                        "field `{}` is not a protobufs primitive, message, map or "
                        "repeated".format(field)
                    ),
                )

        return cls(**kwargs)

    @classmethod
    def from_json(cls, json_str):
        """Build a DataBase from a given JSON string. Use google's protobufs.json_format for
        deserialization

        Args:
            json_str (str or dict): A stringified JSON specification/dict of the
                data_model

        Returns:
            caikit.core.data_model.DataBase: A DataBase object.
        """
        # Get protobufs class required for parsing
        error.type_check("<COR91037250E>", str, dict, json_str=json_str)

        if isinstance(json_str, dict):
            # Convert dict object to a JSON string
            json_str = json.dumps(json_str)

        try:
            # Parse given JSON into google.protobufs.pyext.cpp_message.GeneratedProtocolMessageType
            parsed_proto = json_format.Parse(
                json_str, cls.get_proto_class()(), ignore_unknown_fields=False
            )

            # Use from_proto to return the DataBase object from the parsed proto
            return cls.from_proto(parsed_proto)

        except json_format.ParseError as ex:
            error("<COR90619980E>", ValueError(ex))

    def to_proto(self):
        """Return a new protobufs populated with the information in this data structure."""
        # get the name of the protobufs class
        proto_class = self.__class__.get_proto_class()
        if proto_class is None:
            error(
                "<COR71783827E>",
                AttributeError(
                    "protobufs not found for class `{}`".format(self.__class__)
                ),
            )

        # create the protobufs and call fill_proto to populate it
        return self.fill_proto(proto_class())

    def to_binary_buffer(self):
        """Returns a binary buffer with a serialized protobufs message of this data model"""
        return self.to_proto().SerializeToString()

    def fill_proto(self, proto):
        """Populate a protobufs with the values from this data model object.

        Args:
            proto: A protocol buffer to be populated.
        Returns:
            protobufs: The filled protobufs.

        Notes:
            The protobufs is filled in place, so the argument and the return
            value are the same at the end of this call.
        """

        for field in self.fields:
            try:
                attr = getattr(self, field)

            except AttributeError:
                error(
                    "<COR71783840E>",
                    AttributeError(
                        "class `{}` has no attribute `{}` but it is in the protobufs".format(
                            self.__class__.__name__, field
                        )
                    ),
                )

            if attr is None:
                continue
            if field in self._fields_primitive:
                setattr(proto, field, attr)
            elif field in self._fields_enum:
                if isinstance(attr, Enum):
                    setattr(proto, field, attr.value)
                else:
                    setattr(proto, field, attr)
            elif field in self._fields_map:
                subproto = getattr(proto, field)
                for key, value in attr.items():
                    # If our values aren't primitives, the subproto will have a DESCRIPTOR;
                    # in this case we need to fill down recursively, i.e., this is a
                    # protobufs message map container
                    if hasattr(subproto[key], "DESCRIPTOR"):
                        value.fill_proto(subproto[key])
                    # Otherwise we have a protobufs scalar map container, and we can set the
                    # primitive value like a normal dictionary.
                    else:
                        subproto[key] = value
            elif (
                field in self._fields_primitive_repeated
                or field in self._fields_enum_repeated
            ):
                subproto = getattr(proto, field)
                subproto.extend(attr)

            elif field in self._fields_message:
                subproto = getattr(proto, field)
                if subproto.DESCRIPTOR.full_name == json_dict.STRUCT_PROTOBUF_NAME:
                    subproto.CopyFrom(
                        json_dict.dict_to_struct(attr, subproto.__class__)
                    )
                elif subproto.DESCRIPTOR.full_name == timestamp.TIMESTAMP_PROTOBUF_NAME:
                    timestamp_proto = timestamp.datetime_to_proto(attr)
                    subproto.CopyFrom(timestamp_proto)
                # check that this is any of the Union of List types
                elif subproto.DESCRIPTOR.full_name.endswith(
                    "Sequence"
                ) and not issubclass(attr.__class__, DataBase):
                    seq_dm = subproto.__class__
                    try:
                        subproto.CopyFrom(seq_dm(values=attr))
                        log.debug4("Successfully fill proto for %s", field)
                    except TypeError:
                        log.debug4("not the correct union list type")
                else:
                    attr.fill_proto(subproto)

            elif field in self._fields_message_repeated:
                subproto = getattr(proto, field)
                for item in attr:
                    elem_type = subproto.add()
                    if isinstance(item, dict):
                        elem_type.CopyFrom(
                            json_dict.dict_to_struct(item, elem_type.__class__)
                        )
                    elif isinstance(item, datetime.datetime):
                        elem_type.CopyFrom(timestamp.datetime_to_proto(item))
                    else:
                        item.fill_proto(elem_type)
            else:
                error(
                    "<COR71783852E>",
                    AttributeError(
                        "field `{}` is not a protobufs primitive, message or repeated".format(
                            field
                        )
                    ),
                )

        return proto

    def to_dict(self) -> dict:
        """Convert to a dictionary representation."""
        # maintain a list of fields to convert to dict, special handling for oneofs
        fields_to_dict = []
        for field in self.fields:
            if (
                not field in self._fields_to_oneof
                or self.which_oneof(self._fields_to_oneof[field]) == field
            ):
                fields_to_dict.append(field)

        to_dict = {}
        for field in fields_to_dict:
            dict_value = self._field_to_dict_element(field)
            if (
                field in self._fields_to_oneof
                and not hasattr(dict_value, "values")
                and isinstance(dict_value, list)
            ):
                dict_value = {"values": dict_value}
            to_dict[field] = dict_value
        return to_dict

    def to_kwargs(self) -> dict:
        """Convert to flat dictionary representation. (Like .to_dict, but not recursive)
        This keeps the attribute names of any fields backed by oneofs, instead of using the
        internal oneof field name
        """
        fields_to_dict = []
        for field in self.fields:
            if field not in self._fields_to_oneof:
                fields_to_dict.append(field)
            else:
                fields_to_dict.append(self._fields_to_oneof[field])
        return {field: getattr(self, field) for field in fields_to_dict}

    def to_json(self, **kwargs) -> str:
        """Convert to a json representation."""

        def _default_serialization_overrides(obj):
            """Default handler for nonserializable objects; currently this only handles
            - bytes
            - datetime.datetime
            """
            if isinstance(obj, bytes):
                return base64.b64encode(obj).decode("utf-8")
            if isinstance(obj, datetime.datetime):
                # Use the timestamp's proto-serialized format to get the proper json serializer
                return timestamp.datetime_to_proto(obj).ToJsonString()
            raise TypeError(f"Object of type {type(obj)} is not JSON serializable")

        if "default" not in kwargs:
            kwargs["default"] = _default_serialization_overrides

        return json.dumps(self.to_dict(), **kwargs)

    def __repr__(self):
        """Human-friendly representation."""
        return self.to_json(indent=2, ensure_ascii=False)

    def _field_to_dict_element(self, field):
        """Convert field into a representation that can be placed into a dictionary.  Recursively
        calls to_dict on other data model objects.
        """
        try:
            attr = getattr(self, field)

        except AttributeError:
            error(
                "<COR71783864E>",
                AttributeError(
                    "class `{}` has no attribute `{}` but it is in the protobufs".format(
                        self.__class__.__name__, field
                    )
                ),
            )

        # if field is None, assume it's unset and just return None
        if attr is None:
            return None

        if field in self._fields_enum:
            # if field is an enum, do the reverse lookup from int -> str
            enum_rev = self.fields_enum_rev.get(field)
            if enum_rev is not None:
                return (
                    enum_rev[attr.value] if isinstance(attr, Enum) else enum_rev[attr]
                )

        if field in self._fields_enum_repeated:
            # if field is an enum, do the reverse lookup from int -> str
            enum_rev = self.fields_enum_rev.get(field)
            if enum_rev is not None:
                return [enum_rev[item] for item in attr]

        # if field is a primitive, just return it to be placed in dict
        if field in self._fields_primitive or field in self._fields_primitive_repeated:
            return attr

        def _recursive_to_dict(_attr):
            if isinstance(_attr, dict):
                return {key: _recursive_to_dict(value) for key, value in _attr.items()}
            if isinstance(_attr, list):
                return [_recursive_to_dict(listitem) for listitem in _attr]
            if isinstance(_attr, DataBase):
                return _attr.to_dict()

            return _attr

        # If field is an object in out data model/map/list call to_dict recursively on each element
        if (
            field in self._fields_map
            or field in self._fields_message
            or field in self._fields_message_repeated
        ):
            return _recursive_to_dict(attr)

        # fallback to the string representation
        return str(attr)

    @staticmethod
    def get_class_for_proto(
        proto: Union[Descriptor, ProtoMessageType]
    ) -> Type["DataBase"]:
        """Look up the data model class corresponding to the given protobuf

        If no data model is found, this raises an AttributeError

        Args:
            proto (Union[Descriptor, ProtoMessageType])
                The proto name or descriptor to look up against

        Returns:
            dm_class (Type[DataBase]): The data model class corresponding to the
                given protobuf
        """
        error.type_check(
            "<COR46446770E>",
            Descriptor,
            ProtoMessageType,
            proto=proto,
        )
        proto_full_name = (
            proto.full_name
            if isinstance(proto, Descriptor)
            else proto.DESCRIPTOR.full_name
        )
        cls = _DataBaseMetaClass.class_registry.get(proto_full_name)
        if cls is None:
            error(
                "<COR71783879E>",
                AttributeError(
                    "no data model class found in registry for protobufs named `{}`".format(
                        proto_full_name
                    )
                ),
            )

        return cls

    @staticmethod
    def get_class_for_name(class_name: str) -> Type["DataBase"]:
        """Look up the data model class corresponding to the given name

        This lookup attempts to encode various naming conventions that might be
        used, but it can fail in multiple ways:

        1. No class with the given name is known
        2. Multiple classes with the same name, but different qualified parents
           are found

        A ValueError will be raised if either of the above happens

        Args:
            class_name (str)
                The name of the class either as a fully-qualified protobuf name
                or as the unqualified class name

        Returns:
            dm_class (Type[DataBase]): The data model class corresponding to the
                given protobuf
        """
        dm_class = _DataBaseMetaClass.class_registry.get(class_name)
        if dm_class is not None:
            return dm_class
        matching_classes = [
            (full_name, dm_class)
            for full_name, dm_class in _DataBaseMetaClass.class_registry.items()
            if full_name.rpartition(".")[-1] == class_name
        ]
        if len(matching_classes) == 1:
            return matching_classes[0][1]
        if len(matching_classes) > 1:
            error(
                "<COR02514290E>",
                ValueError(
                    "Conflicting data model classes for [{}]: {}".format(
                        class_name, [match[0] for match in matching_classes]
                    )
                ),
            )
        error(
            "<COR99562895E>",
            ValueError(f"No data model class match for {class_name}"),
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import os
import shutil
import tempfile

# First Party
import alog

# Local
from caikit.runtime.dump_services import dump_grpc_services, dump_http_services

## Helpers #####################################################################

log = alog.use_channel("TEST-DUMP-I")


def test_dump_grpc_services_dir_exists():
    with tempfile.TemporaryDirectory() as workdir:
        dump_grpc_services(workdir)
        assert os.path.exists(workdir)

        for file in os.listdir(workdir):
            assert file.endswith(".proto")


def test_dump_grpc_services_dir_does_not_exist():
    fake_dir = "fake_dir"
    dump_grpc_services(fake_dir)
    assert os.path.exists(fake_dir)

    for file in os.listdir(fake_dir):
        assert file.endswith(".proto")

    shutil.rmtree(fake_dir)


def test_dump_http_services_dir_exists():
    with tempfile.TemporaryDirectory() as workdir:
        dump_http_services(workdir)
        assert os.path.exists(workdir)

        for file in os.listdir(workdir):
            assert file == "openapi.json"
            assert os.path.getsize(os.path.join(workdir, file)) > 0


def test_dump_http_services_dir_does_not_exist():
    fake_dir = "fake_dir"
    dump_http_services(fake_dir)
    assert os.path.exists(fake_dir)

    for file in os.listdir(fake_dir):
        assert file == "openapi.json"
        assert os.path.getsize(os.path.join(fake_dir, file)) > 0

    shutil.rmtree(fake_dir)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Enumeration data structures map from strings to integers and back.
"""

# Standard
from enum import Enum
from typing import Dict, Optional, Tuple, Type

# Third Party
from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper
import google
import munch

# First Party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("DATAM")
error = error_handler.get(log)


@classmethod
def to_dict(cls) -> Dict[str, int]:
    """Return a dict representation of the keys and values"""
    if not hasattr(cls, "__dict_repr__"):
        setattr(
            cls,
            "__dict_repr__",
            {
                entry.name: entry.value
                for entry in cls  # pylint: disable=not-an-iterable
            },
        )
    return cls.__dict_repr__


@classmethod
def to_munch(cls) -> munch.Munch:
    """Return a munchified version of the enum"""
    if not hasattr(cls, "__munch_repr__"):
        setattr(cls, "__munch_repr__", munch.Munch(cls.to_dict()))
    return cls.__munch_repr__


__all__ = ["import_enums", "import_enum"]


def import_enum(
    proto_enum: EnumTypeWrapper, enum_class: Optional[Type[Enum]] = None
) -> Tuple[str, str]:
    """Import a single enum into the global enum module by name

    Args:
        proto_enum (EnumTypeWrapper): The enum to import
        enum_class (Optional[Type[Enum]]): A pre-existing enum class that this
            proto enum binds to

    Returns:
        name:  str
            The name of the enum global
        rev_name:  str
            The name of the reversed enum global
    """
    if not isprotobufenum(proto_enum):
        error(
            "<COR71783964E>",
            AttributeError(f"`{proto_enum}` is not a valid protobuf enumeration"),
        )

    name = proto_enum.DESCRIPTOR.name
    log.debug2("Importing enum named %s", name)
    if enum_class is None:
        log.debug2("Creating Enum class for %s", name)
        enum_class = Enum._create_(name, proto_enum.items())

    # Add extra utility functions
    setattr(enum_class, "to_dict", to_dict)
    setattr(enum_class, "to_munch", to_munch)

    globals()[name] = enum_class
    rev_name = name + "Rev"
    globals()[rev_name] = munch.Munch({v: k for k, v in proto_enum.items()})
    __all__.append(name)
    __all__.append(rev_name)
    return name, rev_name


def import_enums(current_globals):
    """Add all enums and their reverse enum mappings a module's global symbol table. Note that
    we also update __all__. In general, __all__ controls the stuff that comes with a wild (*)
    import.

    Examples tend to make stuff like this easier to understand. Let's say the first name we hit
    is the Entity Mention Type. Then, after the first cycle through the loop below, you'll see
    something like:

        '__all__': ['import_enums', 'EntityMentionType', 'EntityMentionTypeRev']
        'EntityMentionType': { "MENTT_UNSET": 0, "MENTT_NAM": 1, ... , "MENTT_NONE": 4}
        'EntityMentionTypeRev': { "0": "MENTT_UNSET", "1": "MENTT_NAM", ... , "4": "MENTT_NONE"}

    since this is called explicitly below, you can thank this function for automagically syncing
    your enums (as importable from this file) with the data model.

    Args:
        current_globals (dict): global dictionary from your data model package
            __init__ file.
    """
    # Like the proto imports, we'd one day like to do this with introspection using something
    # like below, but can't because our wheel is compiled. If you can think of a cleaner way
    # to do this, open a PR!
    # caller = inspect.stack()[1]
    # caller_module = inspect.getmodule(caller[0])
    # current_globals = caller_module.__dict__

    # Add the str->int (EnumBase) and int->str (EnumRevBase) mapping for each enum
    # to the calling module's symbol table, then update __all__ to include the names
    # for the added objects.
    protobufs = current_globals.get("protobufs")
    all_enum_names = getattr(protobufs, "all_enum_names", [])
    for name in all_enum_names:
        proto_enum = getattr(protobufs, name)
        name, rev_name = import_enum(proto_enum)
        current_globals[name] = globals()[name]
        current_globals[rev_name] = globals()[rev_name]


def isprotobufenum(obj):
    """Returns True if obj is a protobufs enum."""
    return isinstance(obj, google.protobuf.internal.enum_type_wrapper.EnumTypeWrapper)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains functions that attempt to parse the I/O types of member methods on `caikit.core.module`s
"""
# Standard
from typing import Any, Callable, Dict, List, Optional, Type
import inspect

# First Party
import alog

# Local
from ..data_model.base import DataBase
from ..modules import ModuleBase
from . import docstrings

log = alog.use_channel("SIG-PARSING")

# Constants ##################################
KNOWN_ARG_TYPES = {
    "producer_id": "ProducerId",
}


@alog.logged_function(log.debug2)
def get_output_type_name(
    module_class: ModuleBase.__class__,
    fn_signature: inspect.Signature,
    fn: Callable,
) -> Type:
    """Get the type for a return type based on the name of the module class and
    the Caikit library naming convention.
    """
    log.debug(fn_signature)
    # Check type annotation first
    if fn_signature.return_annotation != fn_signature.empty:
        if isinstance(fn_signature.return_annotation, str):
            log.debug("Return type annotation is a string!")
            if fn_signature.return_annotation == module_class.__name__:
                log.debug(
                    "Assuming return annotation is for the module class itself: %s matches %s",
                    module_class,
                    fn_signature.return_annotation,
                )
                return module_class
        else:
            return fn_signature.return_annotation

    # Check the docstring
    type_from_docstring = docstrings.get_return_type(fn)

    if type_from_docstring:
        return type_from_docstring

    # If we get here, it means no annotation or docstring for type was provided

    # Warn unless this was a base function (e.g., don't warn if there is no train() override)
    if fn.__module__ != ModuleBase.__module__:
        log.warning(
            "Could not deduct output type from function %s for module class %s.",
            fn.__name__,
            module_class.__name__,
        )
    else:
        log.debug(
            "Could not deduct output type from function %s for module class %s using %s.",
            fn.__name__,
            module_class.__name__,
            fn.__qualname__,
        )


def get_argument_types(module_method: Callable) -> Dict[str, Type]:
    """Get the python types for each parameter to this method, returned in a dict.
    This does more than simply looking at inspect.Signature, see _get_argument_type

    Args:
        module_method (Callable): A pointer to a method

    Returns:
        Dict[str, Type]: A dictionary of parameter name to parameter type
    """
    method_signature = inspect.signature(module_method)
    return {
        name: _get_argument_type(param, module_method)
        for name, param in method_signature.parameters.items()
        if name not in ["self", "args", "kwargs", "_", "__"]
    }


def get_args_with_defaults(module_method: Callable) -> Dict[str, Any]:
    """Get the the mapping of all argument names that have defaults to their
    default values.

    Args:
        module_method (Callable): A pointer to a method

    Returns:
        Dict[str: Any]: A set of all parameter names which have a default value.
            Empty if none have defaults or no parameters exist.
    """
    method_signature = inspect.signature(module_method)
    return {
        param.name: param.default
        for param in method_signature.parameters.values()
        if param.default != inspect.Parameter.empty
    }


# pylint: disable=too-many-return-statements
@alog.logged_function(log.debug2)
def _get_argument_type(
    arg: inspect.Parameter,
    module_method: Callable,
) -> Type:
    """Get the python type for a named argument to a Module's given method. This
    is where the heuristics for determining types are implemented:

    * Look for a known type mapping based on the name of the argument
    * Look for python type annotations
    * Look for a default value and check its type
    * Parse the docstring
    * Look for a data model object whose name matches the argument name
    """
    # TODO: KNOWN_ARG_TYPES should be configurable

    # Use known arg types first
    # This avoids cases where docstrings are very obviously flubbed, such as
    #   `raw_document` being annotated as a `str` only in caikit.interfaces.nlp
    dm_type_from_known_arg_types = _get_dm_type_from_name(KNOWN_ARG_TYPES.get(arg.name))
    if dm_type_from_known_arg_types:
        # Not checking if this is optional: These known types should never be optional (maybe...?)
        # This could totally be incorrect!
        log.info(
            "Using well known type %s for parameter name %s",
            dm_type_from_known_arg_types,
            arg.name,
        )
        return dm_type_from_known_arg_types

    # Check docstring for optional arg
    optional_arg = docstrings.is_optional(module_method, arg.name)

    # Look for a type annotation
    if arg.annotation != inspect.Parameter.empty:
        log.debug("Found annotation for %s", arg.name)
        if optional_arg:
            return Optional[arg.annotation]
        return arg.annotation

    log.debug("No annotation found for %s", arg.name)

    # Check for a default argument and return its type
    default_type = _get_default_type(arg)
    if default_type:
        return default_type

    # Parse docstring
    type_from_docstring = docstrings.get_arg_type(module_method, arg.name)
    if type_from_docstring:
        if optional_arg:
            return Optional[type_from_docstring]
        return type_from_docstring

    # Look for a data model object whose name matches the argument name and fall
    # back to the KNOWN_ARG_TYPES dict
    candidate_name = _snake_to_camel(arg.name)
    type_from_candidate = _get_dm_type_from_name(candidate_name)
    if optional_arg:
        return Optional[type_from_candidate]
    return type_from_candidate


def _snake_to_camel(string: str) -> str:
    """Simple snake -> camel conversion"""
    return "".join([part[0].upper() + part[1:] for part in string.split("_")])


def _get_dm_type_from_name(data_model_class_name: Optional[str]) -> Type:
    """Given a data model class name, look up the data model class itself"""
    if data_model_class_name is None:
        return None
    try:
        return DataBase.get_class_for_name(data_model_class_name)
    except ValueError:
        return None


def _get_default_type(arg: inspect.Parameter) -> Optional[Type]:
    """
    Tries to infer a type from the default value of the argument
    Args:
        arg: (inspect.Parameter) The inspected argument
    Returns: (Optional[Type]) The type of the argument,: or None if no default value is
        present
    """
    if arg.default != inspect.Parameter.empty and arg.default is not None:
        log.debug3("Found default with type %s", type(arg.default))

        # If the default is a list or tuple, we'll create the corresponding
        # typing type using List. Note that even when the default is a tuple,
        # we use List because we assume that tuple is used to avoid the problem
        # of mutable defaults.
        if isinstance(arg.default, (tuple, list)):
            value_types = {type(val) for val in arg.default}
            if value_types:
                if len(value_types) > 1:
                    log.warning(
                        "Found argument [%s] with iterable default [%s] and multiple types",
                        arg.name,
                        arg.default,
                    )
                return List[list(value_types)[0]]
        else:
            return type(arg.default)
    # Return None if no default argument was given
    return None



"""
A sample module for sample things!
"""
# Standard
from typing import Iterable
import os

# Local
from ...data_model.sample import (
    GeoSpatialTask,
    SampleInputType,
    SampleOutputType,
    SampleTask,
    SampleTrainingType,
)
from caikit.core.data_model import DataStream
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "e0e1e9b1-3cbb-411d-b066-3b8e1a19e46d",
    "WhyAreNamesStillHere",
    "0.0.1",
    GeoSpatialTask,
)
class GeoStreamingModule(caikit.core.ModuleBase):
    @GeoSpatialTask.taskmethod(input_streaming=True, output_streaming=True)
    def run_bidi_stream(
        self, lats: DataStream[float], lons: DataStream[float], name: str
    ) -> Iterable[SampleOutputType]:
        """
        Args:
            sample_inputs (caikit.core.data_model.DataStream[sample_lib.data_model.SampleInputType]): the input

        Returns:
            caikit.core.data_model.DataStream[sample_lib.data_model.SampleOutputType]: The output
                stream
        """
        for lat, lon in DataStream.zip(lats, lons):
            yield SampleOutputType(greeting=f"Hello from {name} at {lat}°, {lon}°")



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Basic routines for reading and writing common file types.
"""

# Standard
import csv
import json
import os
import pickle
import shutil

# Third Party
import yaml


def load_txt(filename):
    """Load a string from a file with utf8 encoding."""
    with open(filename, mode="r", encoding="utf8") as fh:
        return fh.read()


def load_txt_lines(filename):
    """Load a list of files from a text file with utf8 encoding"""
    with open(filename, mode="r", encoding="utf8") as fh:
        wordlist = list(map(str.strip, fh.readlines()))
    return wordlist


def save_txt(text, filename, mode="w"):
    """Write a string to a text file with utf8 encoding."""
    with open(filename, mode=mode, encoding="utf8") as fh:
        fh.write(text)


def load_binary(filename):
    """Load a binary string from a file."""
    with open(filename, mode="rb", encoding="utf-8") as fh:
        return fh.read()


def save_binary(data, filename):
    """Write a binary buffer to a file."""
    with open(filename, mode="wb", encoding="utf-8") as fh:
        fh.write(data)


def load_csv(filename):
    """Load a csv into a list-of-lists."""
    with open(filename, mode="r", newline="", encoding="utf-8") as fh:
        return list(csv.reader(fh, delimiter=",", quotechar='"'))


def save_csv(text_list, filename, mode="w"):
    """Write a list-of-lists to a csv file."""
    with open(filename, mode=mode, newline="", encoding="utf-8") as fh:
        writer = csv.writer(fh, delimiter=",")
        writer.writerows(text_list)


def load_dict_csv(filename):
    """Load a csv into a list-of-dicts."""
    with open(filename, mode="r", encoding="utf-8") as csv_file:
        csv_reader = csv.DictReader(csv_file)
        return list(csv_reader)


def save_dict_csv(dict_list, filename, mode="w"):
    """Write a list of dicts to a csv file."""
    if dict_list:
        keys = dict_list[0].keys()
        with open(filename, mode=mode, encoding="utf-8") as output_file:
            dict_writer = csv.DictWriter(output_file, keys)
            dict_writer.writeheader()
            dict_writer.writerows(dict_list)


def load_json(filename):
    """Load a json file into a dictionary."""
    with open(filename, mode="r", encoding="utf8") as fh:
        return json.load(fh)


def save_json(save_dict, filename, mode="w"):
    """Save a dictionary into a json file."""
    with open(filename, mode=mode, encoding="utf8") as fh:
        json.dump(save_dict, fh, indent=2, ensure_ascii=False)


def load_yaml(filename):
    """Load a yaml file into a dictionary."""
    with open(filename, mode="r", encoding="utf8") as fh:
        return yaml.safe_load(fh)


def save_yaml(save_dict, filename, mode="w"):
    """Save a dictionary into a yaml file."""
    with open(filename, mode=mode, encoding="utf8") as fh:
        yaml.safe_dump(save_dict, fh, default_flow_style=False)


def load_pickle(filename):
    """Load an object from a pickle file."""
    with open(filename, mode="rb") as fh:
        return pickle.load(fh)


def save_pickle(obj, filename, mode="wb"):
    """Save an object to a pickle file."""
    # pylint: disable=unspecified-encoding
    with open(filename, mode=mode) as fh:
        pickle.dump(obj, fh)


def save_raw(save_content, filename, mode="w"):
    """Write the given raw string content to output file."""
    with open(filename, mode=mode, encoding="utf8") as fh:
        fh.write(save_content)


def compress(dir_path, output_path=None, extension="zip"):
    """Compress a given folder recursively to an archive with a given extension format

    Args:
        dir_path (str): Path of directory to compress
        output_path: (Optional) str
            Output path where the archive is created. Defaults to current path + 'archive' +
            format extension
            >>> compress('.', 'my/path', 'tar')
            >>> # saves to 'my/path/archive.tar'

        extension: (Optional) (one of: zip/tar/gztar/bztar/xztar depending on module availability)
            Defaults to .zip

    Returns:
        str: Path to created archive
    """
    if not output_path:
        output_path = os.path.join(os.getcwd(), "archive")

    # Strip away anything preceding '.'
    extension = extension.split(".")[-1]

    shutil.make_archive(output_path, extension, dir_path)
    return output_path + "." + extension



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module holds the common helpers for managing the protobuf package name for
the runtime
"""

# Local
from ...config import get_config


def snake_to_upper_camel(string: str) -> str:
    """Simple snake -> upper camel conversion"""
    return "".join([part[0].upper() + part[1:] for part in string.split("_")])


def get_ai_domain() -> str:
    """Get the string name for the AI domain"""
    caikit_config = get_config()
    lib = caikit_config.runtime.library
    default_ai_domain_name = snake_to_upper_camel(lib.replace("caikit_", ""))
    ai_domain_name = (
        caikit_config.runtime.service_generation.domain or default_ai_domain_name
    )
    return ai_domain_name


def get_runtime_service_package() -> str:
    """This helper will get the common runtime package"""
    caikit_config = get_config()
    ai_domain_name = get_ai_domain()
    default_package_name = f"caikit.runtime.{ai_domain_name}"
    package_name = (
        caikit_config.runtime.service_generation.package or default_package_name
    )
    return package_name



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from . import classification, package, text, text_generation
from .classification import (
    ClassificationResult,
    ClassificationResults,
    ClassificationTrainRecord,
    ClassifiedGeneratedTextResult,
    ClassifiedGeneratedTextStreamResult,
    TokenClassificationResult,
    TokenClassificationResults,
    TokenClassificationStreamResult,
)
from .package import NLP_PACKAGE
from .text import Token, TokenizationResults, TokenizationStreamResult
from .text_generation import (
    FinishReason,
    GeneratedTextResult,
    GeneratedTextStreamResult,
    GeneratedToken,
    TokenStreamDetails,
)
